#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
import binascii
from time import monotonic

from xpra.common import noop
from xpra.util.str_fn import strtobytes

xor_str = noop
try:
    from xpra.buffers import cyxor

    xor_str = cyxor.xor_str
except (ImportError, AttributeError):
    pass


def h(v):
    return binascii.hexlify(v)


class TestHMAC(unittest.TestCase):

    def fail_xor(self, in1, in2):
        try:
            xor_str(in1, in2)
        except Exception:
            pass
        else:
            raise Exception("xor_str did not fail on %s / %s" % (h(in1), h(in2)))

    def check_xor(self, in1, in2, expected):
        out = xor_str(in1, in2)
        #print("xor_str(%s, %s)=%s" % (h(in1), h(in2), h(out)))
        assert out == expected, "expected %s got %s" % (h(expected), h(out))

    def test_xor_str(self):
        for i in range(10):
            l = 16 + i
            zeroes = memoryview(strtobytes(chr(0) * l))
            ones = memoryview(strtobytes(chr(1) * l))
            ff = memoryview(strtobytes(chr(255) * l))
            fe = memoryview(strtobytes(chr(254) * l))
            empty = b""
            lstr = b"\0x80" * 64
            self.check_xor(zeroes, zeroes, zeroes)
            self.check_xor(zeroes[i:], zeroes[i:], zeroes[i:])
            self.check_xor(ones, ones, zeroes)
            self.check_xor(ff, ones, fe)
            self.check_xor(fe, ones, ff)
            self.check_xor(fe[i:], ones[i:], ff[i:])
            #feed some invalid data:
            self.fail_xor(ones, empty)
            self.fail_xor(empty, zeroes)
            self.fail_xor(lstr, ff)
            self.fail_xor(bool, int)

    def test_large_xor_speed(self):
        start = monotonic()
        size = 1 * 1024 * 1024  #1MB
        zeroes = strtobytes(chr(0) * size)
        ones = strtobytes(chr(1) * size)
        count = 10
        for _ in range(count):
            self.check_xor(zeroes, ones, ones)
        end = monotonic()
        if end > start:
            speed = size / (end - start) / 1024 / 1024
            #print("%iMB/s: took %ims on average (%s iterations)" % (speed, 1000*(end-start)/count, count))
            assert speed > 0, "running the xor speed test took too long"


def main():
    if xor_str != noop:
        unittest.main()
    else:
        print("no cyxor module found, test skipped")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time

from unit.client.x11_client_test_util import X11ClientTestUtil
from xpra.util.env import envbool
from xpra.os_util import get_hex_uuid
from xpra.util.io import pollwait, which
from xpra.platform.features import CLIPBOARDS
from xpra.log import Logger

log = Logger("clipboard")

SANITY_CHECKS = envbool("XPRA_CLIPBOARD_SANITY_CHECKS", True)

def has_xclip():
	return bool(which("xclip"))


class X11ClipboardTestUtil(X11ClientTestUtil):

	def get_clipboard_value(self, display, selection="clipboard"):
		out = X11ClientTestUtil.get_command_output("xclip -display %s -selection %s -o" % (display, selection), shell=True)
		return out.decode()

	def set_clipboard_value(self, display, value, selection="clipboard"):
		cmd = "echo -n '%s' | xclip -display %s -selection %s -i" % (value, display, selection)
		xclip = self.run_command(cmd, shell=True)
		assert pollwait(xclip, 5)==0, "xclip command '%s' returned %s" % (cmd, xclip.poll())


	def copy_and_verify(self, display1, display2, synced=True, wait=1, selection="clipboard"):
		log("copy_and_verify%s", (display1, display2, synced, wait, selection))
		value = get_hex_uuid()
		self.set_clipboard_value(display1, value)
		#wait for synchronization to occur:
		time.sleep(wait)
		new_value = self.get_clipboard_value(display2)
		if synced:
			assert new_value==value, "clipboard contents for %s do not match, expected '%s' but got '%s'" % (selection, value, new_value)
		else:
			assert new_value!=value, "clipboard contents for %s match but synchronization was not expected: value='%s'" % (selection, value)
		if SANITY_CHECKS and display2!=display1:
			#verify that the value has not changed on the original display:
			new_value = self.get_clipboard_value(display1)
		return value

	def do_test_copy_selection(self, selection="clipboard", direction="both"):
		log("do_test_copy(%s, %s)", selection, direction)
		server = self.run_server()
		server_display = server.display
		#connect a client:
		xvfb, client = self.run_client(server_display,
									"--clipboard-direction=%s" % direction, "--remote-logging=no")
		assert pollwait(client, 2) is None, "client has exited with return code %s" % client.poll()
		client_display = xvfb.display

		#wait for client to own the clipboard:
		cmd = self.get_xpra_cmd()+["info", server_display]
		for _ in range(10):
			out = X11ClientTestUtil.get_command_output(cmd)
			if out.find(b"clipboard.client=")>0:
				break
			time.sleep(1)

		if SANITY_CHECKS:
			log("sanity checks")
			#xclip sanity check: retrieve from the same display:
			self.copy_and_verify(client_display, client_display, True, wait=0, selection=selection)
			self.copy_and_verify(server_display, server_display, True, wait=0, selection=selection)

		log("copy client %s to server %s", client_display, server_display)
		for _ in range(2):
			self.copy_and_verify(client_display, server_display, direction in ("both", "to-server"), selection=selection)
		log("copy server %s to client %s", server_display, client_display)
		for _ in range(2):
			self.copy_and_verify(server_display, client_display, direction in ("both", "to-client"), selection=selection)
		log("copy client %s to server %s", client_display, server_display)
		for _ in range(2):
			self.copy_and_verify(client_display, server_display, direction in ("both", "to-server"), selection=selection)

		client.terminate()
		xvfb.terminate()
		server.terminate()

	def do_test_copy(self, direction="both"):
		for selection in CLIPBOARDS:
			self.do_test_copy_selection(selection, direction)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import time
from time import monotonic
from multiprocessing import Queue as MQueue, freeze_support
from typing import Any
from collections.abc import Callable

from xpra.util.objects import typedict
from xpra.util.str_fn import csv, repr_ellipsized, print_nested_dict, bytestostr
from xpra.util.env import envint, envbool, envfloat
from xpra.common import ConnectionMessage
from xpra.os_util import (
    get_username_for_uid, get_groups, get_home_for_uid, getuid, getgid, get_group_id, gi_import,
    WIN32, POSIX, OSX,
)
from xpra.util.io import umask_context
from xpra.net.common import PacketType, is_request_allowed
from xpra.net.socket_util import SOCKET_DIR_MODE, SOCKET_DIR_GROUP
from xpra.server import features
from xpra.server.core import ServerCore
from xpra.net.control.common import ArgsControlCommand, ControlError
from xpra.auth.common import SessionData
from xpra.util.child_reaper import getChildReaper
from xpra.scripts.parsing import str_to_bool, MODE_ALIAS
from xpra.scripts.config import make_defaults_struct, PROXY_START_OVERRIDABLE_OPTIONS, OPTION_TYPES
from xpra.scripts.main import parse_display_name, connect_to, start_server_subprocess
from xpra.util.thread import start_thread
from xpra.log import Logger

log = Logger("proxy")
authlog = Logger("proxy", "auth")

GLib = gi_import("GLib")

freeze_support()

PROXY_SOCKET_TIMEOUT = envfloat("XPRA_PROXY_SOCKET_TIMEOUT", 0.1)
PROXY_WS_TIMEOUT = envfloat("XPRA_PROXY_WS_TIMEOUT", 1.0)
assert PROXY_SOCKET_TIMEOUT > 0, "invalid proxy socket timeout"
CAN_STOP_PROXY = envbool("XPRA_CAN_STOP_PROXY", getuid() != 0 or WIN32)
STOP_PROXY_SOCKET_TYPES = os.environ.get("XPRA_STOP_PROXY_SOCKET_TYPES", "named-pipe").split(",")
STOP_PROXY_AUTH_SOCKET_TYPES = os.environ.get("XPRA_STOP_PROXY_AUTH_SOCKET_TYPES", "socket").split(",")
# something (a thread lock?) doesn't allow us to use multiprocessing on MS Windows:
PROXY_INSTANCE_THREADED = envbool("XPRA_PROXY_INSTANCE_THREADED", WIN32)
PROXY_SSL_THREADED = envbool("XPRA_PROXY_SSL_THREADED", False)
PROXY_CLEANUP_GRACE_PERIOD = envfloat("XPRA_PROXY_CLEANUP_GRACE_PERIOD", 0.5)

MAX_CONCURRENT_CONNECTIONS = envint("XPRA_PROXY_MAX_CONCURRENT_CONNECTIONS", 200)
DEFAULT_ENV_WHITELIST: str = "LANG,HOSTNAME,PWD,TERM,SHELL,SHLVL,PATH,USER,HOME"
if WIN32:
    # DEFAULT_ENV_WHITELIST = "ALLUSERSPROFILE,APPDATA,COMMONPROGRAMFILES,COMMONPROGRAMFILES(X86),"+
    #                        "COMMONPROGRAMW6432,COMPUTERNAME,COMSPEC,FP_NO_HOST_CHECK,LOCALAPPDATA,"+
    #                        "NUMBER_OF_PROCESSORS,OS,PATH,PATHEXT,PROCESSOR_ARCHITECTURE,"+
    #                        "PROCESSOR_ARCHITECTURE,PROCESSOR_IDENTIFIER,PROCESSOR_LEVEL,"+
    #                        "PROCESSOR_REVISION,PROGRAMDATA,PROGRAMFILES,PROGRAMFILES(X86),"+
    #                        "PROGRAMW6432,PSMODULEPATH,PUBLIC,SYSTEMDRIVE,SYSTEMROOT,TEMP,TMP,"+
    #                        "USERDOMAIN,WORKGROUP,USERNAME,USERPROFILE,WINDIR,"+
    #                        "XPRA_REDIRECT_OUTPUT,XPRA_LOG_FILENAME,XPRA_ALL_DEBUG"
    DEFAULT_ENV_WHITELIST = "*"
ENV_WHITELIST: list[str] = os.environ.get("XPRA_PROXY_ENV_WHITELIST", DEFAULT_ENV_WHITELIST).split(",")


def get_socktype(proto) -> str:
    try:
        return proto._conn.socktype
    except AttributeError:
        return "unknown"


def create_system_dir(sps: str) -> None:
    if not POSIX or OSX or not sps:
        return
    xpra_group_id = get_group_id(SOCKET_DIR_GROUP)
    if sps.startswith("/run/xpra") or sps.startswith("/var/run/xpra"):
        # create the directory and verify its permissions
        # which should have been set correctly by tmpfiles.d,
        # but may have been set wrong if created by systemd's socket activation instead
        d = sps.split("/xpra")[0] + "/xpra"
        try:
            if os.path.exists(d):
                stat = os.stat(d)
                mode = stat.st_mode
                if (mode & SOCKET_DIR_MODE) != SOCKET_DIR_MODE:
                    log.warn("Warning: invalid permissions on '%s' : %s", d, oct(mode))
                    mode = mode | SOCKET_DIR_MODE
                    log.warn(" changing to %s", oct(mode))
                    os.chmod(d, mode)
                # noinspection PyChainedComparisons
                if xpra_group_id >= 0 and stat.st_gid != xpra_group_id:
                    import grp
                    group = grp.getgrgid(stat.st_gid)[0]
                    log.warn("Warning: invalid group on '%s': %s", d, group)
                    log.warn(" changing to '%s'", SOCKET_DIR_GROUP)
                    os.lchown(d, stat.st_uid, xpra_group_id)
            else:
                log.info("creating '%s' with permissions %s and group '%s'",
                         d, oct(SOCKET_DIR_MODE), SOCKET_DIR_GROUP)
                with umask_context(0):
                    os.mkdir(d, SOCKET_DIR_MODE)
                stat = os.stat(d)
                # noinspection PyChainedComparisons
                if xpra_group_id >= 0 and stat.st_gid != xpra_group_id:
                    os.lchown(d, stat.st_uid, xpra_group_id)
            mode = os.stat(d).st_mode
            log("%s permissions: %s", d, oct(mode))
        except OSError as e:
            log("create_system_dir()", exc_info=True)
            log.error("Error: failed to create or change the permissions on '%s':", d)
            log.estr(e)


def get_proxy_env() -> dict[str, str]:
    env = {k: v for k, v in os.environ.items() if k in ENV_WHITELIST or "*" in ENV_WHITELIST}
    # env var to add to environment of subprocess:
    extra_env_str = os.environ.get("XPRA_PROXY_START_ENV", "")
    if extra_env_str:
        extra_env = {}
        for e in extra_env_str.split(os.path.pathsep):  # ie: "A=1:B=2"
            parts = e.split("=", 1)
            if len(parts) == 2:
                extra_env[parts[0]] = parts[1]
        log("extra_env(%s)=%s", extra_env_str, extra_env)
        env.update(extra_env)
    return env


def get_proxy_server_base_classes() -> tuple[type, ...]:
    classes: list[type] = [ServerCore]
    if features.dbus:
        from xpra.server.mixins.dbus import DbusServer
        classes.append(DbusServer)
    if features.http:
        from xpra.server.mixins.http import HttpServer
        classes.append(HttpServer)
    return tuple(classes)


SERVER_BASES = get_proxy_server_base_classes()
ProxyServerBaseClass = type("ProxyServerBaseClass", SERVER_BASES, {})


class ProxyServer(ProxyServerBaseClass):
    """
        This is the proxy server you can launch with "xpra proxy",
        once authenticated, it will dispatch the connection
        to the session found using the authenticator's
        get_sessions() function.
    """

    def __init__(self):
        log("ProxyServer.__init__()")
        for bc in SERVER_BASES:
            bc.__init__(self)
        self._max_connections = MAX_CONCURRENT_CONNECTIONS
        self._start_sessions = False
        self.session_type = "proxy"
        self.main_loop = None
        # proxy servers may have to connect to remote servers,
        # or even start them, so allow more time before timing out:
        self._accept_timeout += 10
        self.pings = 0
        self._start_sessions = False
        # keep track of the proxy process instances
        # the display they're on and the message queue we can
        # use to communicate with them
        self.instances = {}
        # connections used exclusively for requests:
        self._requests = set()
        self._socket_timeout = PROXY_SOCKET_TIMEOUT
        self._ws_timeout = PROXY_WS_TIMEOUT

    def init(self, opts) -> None:
        log("ProxyServer.init(%s)", opts)
        self.pings = int(opts.pings)
        self._start_sessions = opts.proxy_start_sessions
        for bc in SERVER_BASES:
            bc.init(self, opts)
        # ensure we cache the platform info before intercepting SIGCHLD
        # as this will cause a fork and SIGCHLD to be emitted:
        from xpra.util.version import get_platform_info
        get_platform_info()
        self.child_reaper = getChildReaper()
        create_system_dir(opts.system_proxy_socket)

    def init_control_commands(self) -> None:
        super().init_control_commands()
        if features.control:
            self.add_control_command("stop", ArgsControlCommand("stop",
                                                                "stops the proxy instance on the given display",
                                                                self.handle_stop_command, min_args=1, max_args=1))

    def install_signal_handlers(self, callback: Callable[[int], None]) -> None:
        from xpra.gtk.signals import register_SIGUSR_signals, register_os_signals
        register_os_signals(callback, "Proxy Server")
        register_SIGUSR_signals("Proxy Server")

    def make_dbus_server(self):
        from xpra.server.proxy.proxy_dbus_server import Proxy_DBUS_Server  # pylint: disable=import-outside-toplevel
        return Proxy_DBUS_Server(self)

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        # add shutdown handler
        self._default_packet_handlers["shutdown-server"] = self._process_proxy_shutdown_server

    def _process_proxy_shutdown_server(self, proto, _packet: PacketType) -> None:
        assert proto in self._requests
        self.clean_quit(False)

    def print_screen_info(self) -> None:
        # no screen, we just use a virtual display number
        pass

    def get_server_mode(self) -> str:
        return "proxy"

    def do_run(self) -> None:
        self.main_loop = GLib.MainLoop()
        self.main_loop.run()

    def handle_stop_command(self, *args) -> str:
        display = args[0]
        log("stop command: will try to find proxy process for display %s", display)
        for instance, v in dict(self.instances).items():
            _, disp, _ = v
            if disp == display:
                log.info("stop command: found matching process %s with pid %i for display %s",
                         instance, instance.pid, display)
                self.stop_proxy(instance)
                return f"stopped proxy instance for display {display}"
        raise ControlError(f"no proxy found for display {display}")

    def stop_all_proxies(self, force: bool = False) -> None:
        instances = self.instances
        log("stop_all_proxies() will stop proxy instances: %s", instances)
        for instance in tuple(instances.keys()):
            self.stop_proxy(instance, force)
        log("stop_all_proxies() done")

    def stop_proxy(self, instance, force: bool = False) -> None:
        v = self.instances.get(instance)
        if not v:
            log.error("Error: proxy instance not found for %s", instance)
            return
        log("stop_proxy(%s) is_alive=%s", instance, instance.is_alive())
        if not instance.is_alive() and not force:
            return
        isprocess, _, mq = v
        log("stop_proxy(%s) %s", instance, v)
        # different ways of stopping for process vs threaded implementations:
        if isprocess:
            # send message:
            if force:
                instance.terminate()
            else:
                mq.put_nowait("stop")
                try:
                    mq.close()
                except Exception as e:
                    log("%s() %s", mq.close, e)
        else:
            # direct method call:
            instance.stop(None, "proxy server request")

    def cleanup(self) -> None:
        self.stop_all_proxies()
        for bc in SERVER_BASES:
            bc.cleanup(self)
        start = monotonic()
        live = True
        log("cleanup() proxy instances: %s", self.instances)
        while monotonic() - start < PROXY_CLEANUP_GRACE_PERIOD and live:
            live = tuple(x for x in tuple(self.instances.keys()) if x.is_alive())
            if live:
                log("cleanup() still %i proxies alive: %s", len(live), live)
                time.sleep(0.1)
        if live:
            self.stop_all_proxies(True)
        log("cleanup() frames remaining:")
        from xpra.util.pysystem import dump_all_frames
        dump_all_frames(log)

    def do_quit(self) -> None:
        self.main_loop.quit()
        # from now on, we can't rely on the main loop:
        from xpra.util.system import register_SIGUSR_signals
        register_SIGUSR_signals()

    def log_closing_message(self) -> None:
        log.info("Proxy Server process ended")

    def verify_connection_accepted(self, protocol) -> None:
        # if we start a proxy, the protocol will be closed
        # (a new one is created in the proxy process)
        if not protocol.is_closed():
            self.send_disconnect(protocol, ConnectionMessage.LOGIN_TIMEOUT)

    def hello_oked(self, proto, c: typedict, auth_caps: dict) -> None:
        if super().hello_oked(proto, c, auth_caps):
            # already handled in superclass
            return
        self.accept_client(proto, c)
        request = c.strget("request")

        if request:
            if not self.handle_hello_request(request, proto, c):
                self.send_disconnect(proto, f"error: invalid request, {request!r} is not supported by the proxy server")
            return
        self.proxy_auth(proto, c, auth_caps)

    def handle_hello_request(self, request: str, proto, caps: typedict) -> bool:
        if request == "stop":
            self.handle_stop_request(proto)
            return True
        if request == "id":
            self.send_id_info(proto)
            return True
        if request == "info":
            info = self.get_session_id_info()
            log_file = os.environ.get("XPRA_SERVER_LOG", "")
            info["log-file"] = log_file
            self.do_send_info(proto, info)
            return True
        return False

    def handle_stop_request(self, proto) -> None:
        socktype = get_socktype(proto)
        default_can_stop = CAN_STOP_PROXY
        default_can_stop |= socktype in STOP_PROXY_SOCKET_TYPES
        default_can_stop |= (socktype in STOP_PROXY_AUTH_SOCKET_TYPES) and proto.authenticators
        if not is_request_allowed(proto, "stop", default_can_stop):
            msg = "`stop` requests are not enabled for this proxy server connection"
            log.warn(f"Warning: {msg}")
            self.send_disconnect(proto, msg)
            return

        self._requests.add(proto)
        # send a hello back and the client should then send its "shutdown-server" packet
        capabilities = self.make_hello(None)
        proto.send_now(("hello", capabilities))

        def force_exit_request_client() -> None:
            try:
                self._requests.remove(proto)
            except KeyError:
                pass
            if not proto.is_closed():
                self.send_disconnect(proto, "timeout")

        GLib.timeout_add(10 * 1000, force_exit_request_client)

    def proxy_auth(self, client_proto, c: typedict, auth_caps: dict) -> None:
        def disconnect(reason, *extras) -> None:
            log("disconnect(%s, %s)", reason, extras)
            self.send_disconnect(client_proto, reason, *extras)

        def nosession(*extras) -> None:
            disconnect(ConnectionMessage.SESSION_NOT_FOUND, *extras)

        # find the target server session:
        if not client_proto.authenticators:
            log.error("Error: the proxy server requires an authentication mode,")
            try:
                log.error(" client connection '%s' does not specify one", get_socktype(client_proto))
            except AttributeError:
                pass
            log.error(" use 'none' to disable authentication")
            nosession("no sessions found")
            return
        sessions : SessionData | None = None
        for authenticator in client_proto.authenticators:
            try:
                auth_sessions = authenticator.get_sessions()
                authlog("proxy_auth %s.get_sessions()=%s", authenticator, auth_sessions)
                if auth_sessions:
                    sessions = auth_sessions
                    break
            except Exception as e:
                authlog("failed to get the list of sessions from %s", authenticator, exc_info=True)
                authlog.error("Error: failed to get the list of sessions using '%s' authenticator", authenticator)
                authlog.estr(e)
                disconnect(ConnectionMessage.AUTHENTICATION_ERROR, "cannot access sessions")
                return
        authlog("proxy_auth(%s, {..}, %s) found sessions: %s", client_proto, auth_caps, sessions)
        if sessions is None:
            nosession("no sessions found")
            return
        self.proxy_session(client_proto, c, auth_caps, sessions)

    def proxy_session(self, client_proto, c: typedict, auth_caps: dict, sessions: SessionData) -> None:
        def disconnect(reason, *extras) -> None:
            log("disconnect(%s, %s)", reason, extras)
            self.send_disconnect(client_proto, reason, *extras)

        def nosession(*extras) -> None:
            disconnect(ConnectionMessage.SESSION_NOT_FOUND, *extras)

        uid, gid, displays, env_options, session_options = sessions
        if POSIX:
            if getuid() == 0:
                if uid == 0 or gid == 0:
                    log.error("Error: proxy instances cannot run as root")
                    log.error(" use a different uid and gid (ie: nobody)")
                    disconnect(ConnectionMessage.AUTHENTICATION_ERROR, "cannot run proxy instances as root")
                    return
            else:
                uid = getuid()
                gid = getgid()
            username = get_username_for_uid(uid)
            password = None
            groups = get_groups(username)
            log("username(%i)=%s, groups=%s", uid, username, groups)
        else:
            # the auth module recorded the username we authenticate against
            assert client_proto.authenticators
            username = password = ""
            for authenticator in client_proto.authenticators:
                username = getattr(authenticator, "username", "")
                password = authenticator.get_password()
                if username:
                    break
        # ensure we don't loop back to the proxy:
        proxy_virtual_display = os.environ.get("DISPLAY")
        if proxy_virtual_display in displays:
            displays.remove(proxy_virtual_display)
        # remove proxy instance virtual displays:
        displays = [x for x in displays if not x.startswith(":proxy-")]
        # log("unused options: %s, %s", env_options, session_options)
        proc = None
        socket_path = None
        display = None
        sns = c.dictget("start-new-session", {})
        authlog("proxy_session: displays=%s, start_sessions=%s, start-new-session=%s",
                displays, self._start_sessions, sns)
        if not displays or sns:
            if not self._start_sessions:
                nosession("no displays found")
                return
            try:
                proc, socket_path, display = self.start_new_session(username, password, uid, gid, sns, displays)
                log("start_new_session%s=%s", (username, "..", uid, gid, sns, displays), (proc, socket_path, display))
            except Exception as e:
                log("start_server_subprocess failed", exc_info=True)
                log.error("Error: failed to start new session subprocess:")
                log.estr(e)
                nosession("failed to start a new session")
                return
        if display is None:
            display = c.strget("display")
            authlog("proxy_session: proxy-virtual-display=%s (ignored), user specified display=%s, found displays=%s",
                    proxy_virtual_display, display, displays)
            if display == proxy_virtual_display:
                nosession("invalid display: proxy display")
                return
            if display:
                if display not in displays:
                    if f":{display}" in displays:
                        display = f":{display}"
                    else:
                        nosession(f"display {display!r} not found")
                        return
            else:
                if len(displays) != 1:
                    nosession("please specify a display, more than one is available: " + csv(sorted(displays)))
                    return
                display = displays[0]

        connect = c.boolget("connect", True)
        # ConnectTestXpraClient doesn't want to connect to the real session either:
        ctr = c.strget("connect_test_request")
        log("connect=%s, connect_test_request=%s", connect, ctr)
        if not connect or ctr:
            log("proxy_session: not connecting to the session")
            hello = {"display": display}
            if socket_path:
                hello["socket-path"] = socket_path
            # echo mode if present:
            mode = str(sns.get("mode", ""))
            if mode:
                hello["mode"] = mode
            client_proto.send_now(("hello", hello))
            return

        def stop_server_subprocess() -> None:
            log("stop_server_subprocess() proc=%s", proc)
            if proc and proc.poll() is None:
                proc.terminate()

        log("start_proxy(%s, {..}, %s) using server display at: %s", client_proto, auth_caps, display)

        def parse_error(*args) -> None:
            stop_server_subprocess()
            nosession("invalid display string")
            log.warn("Error: parsing failed for display string '%s':", display)
            for arg in args:
                log.warn(" %s", arg)
            raise ValueError(f"parse error on {display!r} {args}")

        opts = make_defaults_struct(username=username, uid=uid, gid=gid)
        opts.username = username
        disp_desc = parse_display_name(parse_error, opts, display)
        if uid or gid:
            disp_desc["uid"] = uid
            disp_desc["gid"] = gid
        log("display description(%s) = %s", display, disp_desc)
        try:
            server_conn = connect_to(disp_desc, opts)
        except Exception as e:
            log("cannot connect", exc_info=True)
            log.error("Error: cannot start proxy connection:")
            for x in str(e).splitlines():
                log.error(" %s", x)
            log.error(" connection definition:")
            print_nested_dict(disp_desc, prefix=" ", lchar="*", pad=20, print_fn=log.error)
            nosession("failed to connect to display")
            stop_server_subprocess()
            return
        log("server connection=%s", server_conn)

        cipher = cipher_mode = ""
        encryption_key = b""
        if auth_caps:
            cipher = auth_caps.get("cipher")
            if cipher:
                from xpra.net.crypto import DEFAULT_MODE  # pylint: disable=import-outside-toplevel
                cipher_mode = auth_caps.get("cipher.mode", DEFAULT_MODE)
                encryption_key = self.get_encryption_key(client_proto.authenticators, client_proto.keyfile)

        use_thread = PROXY_INSTANCE_THREADED
        if not use_thread and PROXY_SSL_THREADED:
            client_socktype = get_socktype(client_proto)
            server_socktype = disp_desc["type"]
            if client_socktype in ("ssl", "wss", "ssh"):
                log.info("using threaded mode for %s client connection", client_socktype)
                use_thread = True
            elif server_socktype in ("ssl", "wss", "ssh"):
                log.info("using threaded mode for %s server connection", server_socktype)
                use_thread = True
        if use_thread:
            if env_options:
                log.warn("environment options are ignored in threaded mode")
            from xpra.server.proxy.instance_thread import ProxyInstanceThread
            pit = ProxyInstanceThread(session_options, self.pings,
                                      client_proto, server_conn,
                                      disp_desc, cipher, cipher_mode, encryption_key, c)
            pit.stopped = self.reap
            pit.run()
            self.instances[pit] = (False, display, None)
            return

        # this may block, so run it in a thread:
        def start_proxy_process() -> None:
            log("start_proxy_process()")
            message_queue = MQueue()
            client_conn = None
            try:
                # no other packets should be arriving until the proxy instance responds to the initial hello packet
                def unexpected_packet(packet) -> None:
                    if packet:
                        log.warn("Warning: received an unexpected packet")
                        log.warn(" from the proxy connection %s:", client_proto)
                        log.warn(" %s", repr_ellipsized(packet))
                        client_proto.close()

                client_conn = client_proto.steal_connection(unexpected_packet)
                client_state = client_proto.save_state()
                log("start_proxy_process(..) client connection=%s", client_conn)
                log("start_proxy_process(..) client state=%s", client_state)

                ioe = client_proto.wait_for_io_threads_exit(5 + self._socket_timeout)
                if not ioe:
                    log.error("Error: some network IO threads have failed to terminate")
                    client_proto.close()
                    return
                client_conn.set_active(True)
                from xpra.server.proxy.instance_process import ProxyInstanceProcess
                process = ProxyInstanceProcess(uid, gid, env_options, session_options, self._socket_dir,
                                               self.pings,
                                               client_conn, disp_desc, client_state,
                                               cipher, cipher_mode, encryption_key, server_conn, c, message_queue)
                log("starting %s from pid=%s", process, os.getpid())
                self.instances[process] = (True, display, message_queue)
                process.start()
                log("ProxyInstanceProcess started")
                popen = process._popen
                assert popen
                # when this process dies, run reap to update our list of proxy instances:
                self.child_reaper.add_process(popen, f"xpra-proxy-{display}",
                                              "xpra-proxy-instance", True, True, self.reap)
            except Exception as pie:
                log("start_proxy_process() failed", exc_info=True)
                log.error("Error starting proxy instance process:")
                log.estr(pie)
                message_queue.put(f"error: {pie}")
                message_queue.put("stop")
            finally:
                # now we can close our handle on the connection:
                log("handover complete: closing connection from proxy server")
                if client_conn:
                    client_conn.close()
                server_conn.close()
                log("sending socket-handover-complete")
                message_queue.put("socket-handover-complete")

        start_thread(start_proxy_process, f"start_proxy({client_proto})")

    def start_new_session(self, username: str, _password, uid: int, gid: int,
                          sess_options: dict, displays=()) -> tuple[Any, str, str]:
        log("start_new_session%s", (username, "..", uid, gid, sess_options, displays))
        sns = typedict(sess_options or {})
        mode = sns.strget("mode", "start")
        mode = MODE_ALIAS.get(mode, mode)
        if mode not in ("seamless", "desktop", "shadow", "monitor", "expand"):
            raise ValueError(f"invalid start-new-session mode {mode!r}")
        display = sns.strget("display")
        if display in displays:
            raise ValueError(f"display {display} is already active!")
        log("starting new server subprocess: mode=%s, display=%s", mode, display)
        args = []
        if display:
            args = [display]
        # allow the client to override some options:
        opts = make_defaults_struct(username=username, uid=uid, gid=gid)
        for k, v in sns.items():
            k = bytestostr(k)
            if k in ("mode", "display"):
                continue  # those special attributes have been consumed already
            if k not in PROXY_START_OVERRIDABLE_OPTIONS:
                log.warn("Warning: ignoring invalid start override")
                log.warn(" %s=%s", k, v)
                continue
            vt = OPTION_TYPES[k]
            try:
                if vt is str:
                    v = bytestostr(v)
                elif vt is bool:
                    v = str_to_bool(v)
                elif vt is int:
                    v = int(v)
                elif vt is list:
                    v = list(bytestostr(x) for x in v)
            except ValueError:
                log("start_new_session: override option %s", k, exc_info=True)
                log.warn("Warning: ignoring invalid value %s for %s (%s)", v, k, vt)
                continue
            if v is not None:
                fn = k.replace("-", "_")
                curr = getattr(opts, fn, None)
                if curr != v:
                    log("start override: %24s=%-24s (default=%s)", k, v, curr)
                    setattr(opts, fn, v)
                else:
                    log("start override: %24s=%-24s (unchanged)", k, v)
            else:
                log("start override: %24s=%-24s (invalid, unchanged)", k, v)
        opts.attach = False
        opts.start_via_proxy = False
        env = get_proxy_env()
        cwd = None
        if uid > 0:
            cwd = get_home_for_uid(uid) or None
            if not cwd or not os.path.exists(cwd):
                import tempfile
                cwd = tempfile.gettempdir()
        log("starting new server subprocess: options=%s", opts)
        log("env=%s", env)
        log("args=%s", args)
        log("cwd=%s", cwd)
        proc, socket_path, display = start_server_subprocess(sys.argv[0], args,
                                                             mode, opts, username, uid, gid, env, cwd)
        if proc:
            self.child_reaper.add_process(proc, "server-%s" % (display or socket_path), f"xpra {mode}", True, True)
        log("start_new_session(..) pid=%s, socket_path=%s, display=%s, ", proc.pid, socket_path, display)
        return proc, socket_path, display

    def reap(self, *args) -> None:
        log("reap%s", args)
        dead = []
        for instance in tuple(self.instances.keys()):
            # instance is a process
            if not instance.is_alive():
                dead.append(instance)
        log("reap%s dead processes: %s", args, dead or None)
        for p in dead:
            del self.instances[p]

    def get_info(self, proto, *_args) -> dict[str, Any]:
        authenticated = proto and proto.authenticators
        if not authenticated:
            info = self.get_minimal_server_info()
        else:
            # only show more info if we have authenticated
            # as the user running the proxy server process:
            info = super().get_info(proto)
            sessions = ()
            for authenticator in proto.authenticators:
                auth_sessions = authenticator.get_sessions()
                if auth_sessions:
                    sessions = auth_sessions
                    break
            if sessions:
                uid, gid = sessions[:2]
                if not POSIX or (uid == getuid() and gid == getgid()):
                    self.reap()
                    i = 0
                    instances = dict(self.instances)
                    instances_info = {}
                    for proxy_instance, v in instances.items():
                        isprocess, d, _ = v
                        iinfo = {
                            "display": d,
                            "live": proxy_instance.is_alive(),
                        }
                        if isprocess:
                            iinfo.update({
                                "pid": proxy_instance.pid,
                            })
                        else:
                            iinfo.update(proxy_instance.get_info())
                        instances_info[i] = iinfo
                        i += 1
                    info["instances"] = instances_info
                    info["proxies"] = len(instances)
        info.setdefault("server", {})["type"] = "Python/GLib/proxy"
        return info



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

if sys.platform.startswith("win") or sys.platform == "darwin":    # pragma: no cover
    raise ImportError("no X11 support on %s" % sys.platform)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

"""
Intercepting thread creation

These wrapper functions are here so that we can more easily intercept
the creation of all daemon threads and inject some code.

This is used by the `pycallgraph` test wrapper.
(this is cleaner than overriding the threading module directly
 as only our code will be affected)
"""

import threading
from threading import Thread
from collections.abc import Callable


def make_thread(target: Callable, name: str, daemon: bool = False, args=()) -> Thread:
    t = Thread(target=target, name=name, args=args)
    t.daemon = daemon
    return t


def start_thread(target: Callable, name: str, daemon: bool = False, args=()) -> Thread:
    t = make_thread(target, name, daemon, args=args)
    t.start()
    return t


main_thread = threading.current_thread()


def set_main_thread(thread=threading.current_thread()):
    global main_thread
    main_thread = thread


def is_main_thread() -> bool:
    return threading.current_thread() == main_thread



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any

from xpra.net.common import PacketType
from xpra.util.objects import typedict
from xpra.util.env import SilenceWarningsContext
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

log = Logger("cursor")


class CursorManager(StubServerMixin):
    """
    Mixin for servers that handle cursors.
    """
    PREFIX = "cursor"

    def __init__(self):
        self.cursors = False
        self.cursor_size = 0

    def init(self, opts) -> None:
        self.cursors = opts.cursors

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        if not send_ui:
            return
        if share_count > 0:
            self.cursor_size = 24
        else:
            self.cursor_size = c.intget("cursor.size", 0)

    def send_initial_data(self, ss, caps, send_ui: bool, share_count: int) -> None:
        if not send_ui:
            return
        self.send_initial_cursors(ss, share_count > 0)

    def send_initial_cursors(self, ss, sharing=False) -> None:
        log("send_initial_cursors(%s, %s)", ss, sharing)
        from xpra.server.source.cursors import CursorsMixin
        if isinstance(ss, CursorsMixin):
            ss.send_cursor()

    def get_caps(self, source) -> dict[str, Any]:
        caps = {
            "cursors": self.cursors,
        }
        Gdk = sys.modules.get("gi.repository.Gdk", None)
        display = Gdk.Display.get_default() if Gdk else None
        if display:
            max_size = tuple(display.get_maximal_cursor_size())
            caps["cursor"] = {
                "default_size": display.get_default_cursor_size(),
                "max_size": max_size,
            }
        return caps

    def get_info(self, _proto) -> dict[str, Any]:
        return {
            CursorManager.PREFIX: {
                "": self.cursors,
                "size": self.cursor_size,
            },
        }

    def get_ui_info(self, _proto, _client_uuids=None, *args) -> dict[str, Any]:
        # (from UI thread)
        # now cursor size info:
        Gdk = sys.modules.get("gi.repository.Gdk", None)
        if not Gdk:
            return {}
        display = Gdk.Display.get_default()
        if not display:
            return {}
        with SilenceWarningsContext(DeprecationWarning):
            pos = display.get_default_screen().get_root_window().get_pointer()
        cinfo = {"position": (pos.x, pos.y)}
        for prop, size in {
            "default": display.get_default_cursor_size(),
            "max": tuple(display.get_maximal_cursor_size()),
        }.items():
            if size is None:
                continue
            cinfo[f"{prop}_size"] = size
        return {CursorManager.PREFIX: cinfo}

    def _process_set_cursors(self, proto, packet: PacketType) -> None:
        self._process_cursor_set(proto, packet)

    def _process_cursor_set(self, proto, packet: PacketType) -> None:
        assert self.cursors, "cannot toggle send_cursors: the feature is disabled"
        ss = self.get_server_source(proto)
        if ss:
            ss.send_cursors = bool(packet[1])

    def init_packet_handlers(self) -> None:
        self.add_packets(f"{CursorManager.PREFIX}-set")
        self.add_legacy_alias("set-cursors", f"{CursorManager.PREFIX}-set")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from Foundation import NSUserNotificationCenter, NSUserNotification, NSUserNotificationDefaultSoundName

from xpra.notifications.notifier_base import NotifierBase, NID
from xpra.platform.darwin.gui import log, GLib, notifylog
from xpra.util.env import envbool


class OSX_Notifier(NotifierBase):

    def __init__(self, closed_cb=None, action_cb=None):
        super().__init__(closed_cb, action_cb)
        self.gtk_notifier = None
        self.gtk_notifications = set()
        self.notifications = {}
        self.notification_center = NSUserNotificationCenter.defaultUserNotificationCenter()
        assert self.notification_center

    def show_notify(self, dbus_id, tray, nid: NID,
                    app_name: str, replaces_nid: NID, app_icon,
                    summary: str, body: str, actions, hints, expire_timeout: int, icon):
        GTK_NOTIFIER = envbool("XPRA_OSX_GTK_NOTIFIER", True)
        if actions and GTK_NOTIFIER:
            # try to use GTK notifier if we have actions buttons to handle:
            try:
                from xpra.gtk.notifier import GTKNotifier
            except ImportError as e:
                log("cannot use GTK notifier for handling actions: %s", e)
            else:
                self.gtk_notifier = GTKNotifier(self.closed_cb, self.action_cb)
                self.gtk_notifier.show_notify(dbus_id, tray, nid, app_name, replaces_nid, app_icon,
                                              summary, body, actions, hints, expire_timeout, icon)
                self.gtk_notifications.add(nid)
                return
        GLib.idle_add(self.do_show_notify, dbus_id, tray, nid, app_name, replaces_nid, app_icon, summary, body, actions,
                      hints, expire_timeout, icon)

    def do_show_notify(self, dbus_id, tray, nid: NID, app_name: str,
                       replaces_nid: NID, app_icon,
                       summary: str, body: str, actions, hints, expire_timeout: int, icon):
        notification = NSUserNotification.alloc()
        notification.init()
        notification.setTitle_(summary)
        notification.setInformativeText_(body)
        notification.setIdentifier_("%s" % nid)
        # enable sound:
        notification.setSoundName_(NSUserNotificationDefaultSoundName)
        notifylog("do_show_notify(..) nid=%s, %s(%s)", nid, self.notification_center.deliverNotification_, notification)
        self.notifications[int(nid)] = notification
        self.notification_center.deliverNotification_(notification)

    def close_notify(self, nid: NID):
        try:
            self.gtk_notifications.remove(int(nid))
        except KeyError:
            notification = self.notifications.get(int(nid))
            notifylog("close_notify(..) notification[%i]=%s", nid, notification)
            if notification:
                GLib.idle_add(self.notification_center.removeDeliveredNotification_, notification)
        else:
            self.gtk_notifier.close_notify(int(nid))

    def cleanup(self):
        super().cleanup()
        GLib.idle_add(self.notification_center.removeAllDeliveredNotifications)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#pylint: disable=line-too-long

import unittest

from unit.test_util import LoggerSilencer
from xpra.audio import common


class TestCommon(unittest.TestCase):

    def test_audio_option(self):
        ONE_OPTION = (common.VORBIS, )
        assert common.audio_option_or_all("unspecified", None, ONE_OPTION)==ONE_OPTION
        assert common.audio_option_or_all("unspecified", (), ONE_OPTION)==ONE_OPTION

        assert common.audio_option_or_all("valid", ONE_OPTION, ONE_OPTION)==ONE_OPTION
        with LoggerSilencer(common):
            assert common.audio_option_or_all("invalid options", (common.VORBIS, common.OGG, ), ONE_OPTION)==ONE_OPTION
            assert common.audio_option_or_all("no valid options", (common.VORBIS, common.OGG, ), ())==()


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time
import unittest

from xpra.util.objects import AdHocStruct
from xpra.os_util import get_hex_uuid
from xpra.client.mixins.network_state import NetworkState
from unit.client.mixins.clientmixintest_util import ClientMixinTest


class MixinsTest(ClientMixinTest):

    def test_networkstate(self):
        opts = AdHocStruct()
        opts.pings = True
        opts.bandwidth_limit = 0
        opts.bandwidth_detection = True
        self._test_mixin_class(NetworkState, opts, {"start_time": time.time()})
        x = self.mixin
        if x.server_start_time < x.start_time:
            raise Exception("invalid time: server=%s vs start=%s" % (x.server_start_time, x.start_time))
        x.uuid = get_hex_uuid()
        x.send_info_request()
        packet = ["info-response", {"foo": "bar"}]
        self.handle_packet(packet)
        assert x.server_last_info.get("foo") == "bar"


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import shlex
import unittest
from time import sleep

from xpra.os_util import OSX, POSIX
from xpra.util.io import pollwait
from xpra.util.str_fn import strtobytes
from xpra.exit_codes import ExitCode
from unit.server_test_util import ServerTestUtil, estr, log


class ServerAuthTest(ServerTestUtil):

    @classmethod
    def setUpClass(cls):
        ServerTestUtil.setUpClass()
        cls.xvfb = cls.start_Xvfb()
        cls.display = cls.xvfb.display

    @classmethod
    def tearDownClass(cls):
        cls.xvfb.terminate()
        ServerTestUtil.tearDownClass()

    def _test_auth(self, auth="fail", uri_prefix="", exit_code=0, password=None):
        display = self.xvfb.display
        log("starting test server on %s", display)
        server_args = ["--bind=noabstract", f"--auth={auth}", "--use-display=yes"]
        server = self.check_fast_start_server(display, *server_args)
        # we should always be able to get the version:
        display_uri = uri_prefix + display
        client = self.run_xpra(["version", display_uri])
        assert pollwait(client, 5) == 0, f"version client failed to connect using {display_uri}"
        if client.poll() is None:
            client.terminate()
        # try to connect
        cmd = ["info", display_uri]
        f = None
        try:
            if password:
                f = self._temp_file(strtobytes(password))
                filename = os.path.abspath(f.name)
                cmd += [f"--password-file={filename}"]
                cmd += [f"--challenge-handlers=file:filename={filename}"]
            client = self.run_xpra(cmd)
            r = pollwait(client, 5)
        finally:
            if f:
                f.close()
                self.delete_temp_file(f)

        exit_code_error = r != exit_code
        if exit_code_error:
            log.warn("Server Process Pipes:")
            log.warn(f" for server args={server_args}")
            self.show_proc_pipes(server)
            log.warn("Client Process Pipes:")
            log.warn(f" for client args={cmd}")
            self.show_proc_pipes(client)

        if client.poll() is None:
            client.terminate()

        try:
            server.terminate()
        finally:
            sleep(2)
            self.run_xpra(["clean-sockets"])
        if exit_code_error:
            raise RuntimeError(f"{auth!r} test error: expected info client to return {estr(exit_code)}"
                               f" but got {estr(r)}"
                               f" for server with args={shlex.join(server_args)}"
                               f" and client command: {shlex.join(cmd)}")

    def test_fail(self):
        self._test_auth("fail", "", ExitCode.CONNECTION_FAILED)

    def test_reject(self):
        self._test_auth("reject", "", ExitCode.PASSWORD_REQUIRED)

    def test_none(self):
        self._test_auth("none", "", ExitCode.OK)
        self._test_auth("none", "", ExitCode.OK, "foo")

    def test_allow(self):
        self._test_auth("allow", "", ExitCode.PASSWORD_REQUIRED)
        self._test_auth("allow", "", ExitCode.OK, "foo")

    def test_file(self):
        from xpra.os_util import get_hex_uuid
        password = get_hex_uuid()
        f = self._temp_file(strtobytes(password))
        filename = os.path.abspath(f.name)
        try:
            self._test_auth("file", "", ExitCode.PASSWORD_REQUIRED)
            self._test_auth(f"file:filename={filename}", "", ExitCode.PASSWORD_REQUIRED)
            self._test_auth(f"file:filename={filename}", "", ExitCode.OK, password)
            self._test_auth(f"file:filename={filename}", "", ExitCode.AUTHENTICATION_FAILED, password+"A")
        finally:
            f.close()

    def test_multifile(self):
        from xpra.platform.info import get_username
        username = get_username()
        from xpra.os_util import get_hex_uuid
        password = get_hex_uuid()
        displays = ""
        data = "%s|%s|%i|%i|%s||" % (username, password, os.getuid(), os.getgid(), displays)
        f = self._temp_file(strtobytes(data))
        filename = os.path.abspath(f.name)
        try:
            self._test_auth("multifile", "", ExitCode.PASSWORD_REQUIRED)
            self._test_auth(f"multifile:filename={filename}", "", ExitCode.PASSWORD_REQUIRED)
            self._test_auth(f"multifile:filename={filename}", "", ExitCode.OK, password)
            self._test_auth(f"multifile:filename={filename}", "", ExitCode.AUTHENTICATION_FAILED, password+"A")
        finally:
            f.close()


def main():
    if POSIX and not OSX:
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.util.str_fn import Ellipsizer
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.keyboard.mask import DEFAULT_MODIFIER_MEANINGS
from xpra.util.objects import typedict
from xpra.log import Logger

log = Logger("keyboard")
ibuslog = Logger("ibus")


class InputMixin(StubSourceMixin):
    """
    Manage input devices (keyboard, mouse, etc)
    """

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        return any((
            caps.boolget("keyboard"),
            caps.boolget("mouse"),
            bool(caps.get("xkbmap_keycodes")),  # legacy clients
        ))

    def init_state(self) -> None:
        self.keyboard_config = None
        self.ibus = False
        self.double_click_time: int = -1
        self.double_click_distance: tuple[int, int] | None = None
        # mouse echo:
        self.mouse_last_position: tuple[int, int] | None = None
        self.mouse_last_relative_position: tuple[int, int] | None = None

    def cleanup(self) -> None:
        self.keyboard_config = None

    def parse_client_caps(self, c: typedict) -> None:
        dc = c.dictget("double_click")
        if dc:
            dc = typedict(dc)
            self.double_click_time = dc.intget("time")
            self.double_click_distance = dc.intpair("distance")
        else:
            self.double_click_time = c.intget("double_click.time")
            self.double_click_distance = c.intpair("double_click.distance")
        self.mouse_last_position = c.intpair("mouse.initial-position")
        self.ibus = c.boolget("ibus")
        ibuslog(f"client ibus support: {self.ibus}")

    def get_info(self) -> dict[str, Any]:
        dc_info: dict[str, Any] = {}
        dct = self.double_click_time
        if dct:
            dc_info["time"] = dct
        dcd = self.double_click_distance
        if dcd:
            dc_info["distance"] = dcd
        info = {}
        if dc_info:
            info["double-click"] = dc_info
        kc = self.keyboard_config
        if kc:
            info["keyboard"] = kc.get_info()
        return info

    def get_caps(self) -> dict[str, Any]:
        caps: dict[str, Any] = {}
        # expose the "modifier_client_keycodes" defined in the X11 server keyboard config object,
        # so clients can figure out which modifiers map to which keys:
        mck = getattr(self.keyboard_config, "modifier_client_keycodes", None)
        if mck:
            caps["modifier_keycodes"] = mck
        return caps

    def send_ibus_layouts(self, ibus_layouts: dict) -> None:
        ibuslog(f"send_ibus_layouts(%s) {self.ibus=}", Ellipsizer(ibus_layouts))
        if self.ibus and ibus_layouts:
            self.send_setting_change("ibus-layouts", ibus_layouts)

    def set_layout(self, layout: str, variant: str, options) -> bool:
        if not self.keyboard_config:
            return False
        return self.keyboard_config.set_layout(layout, variant, options)

    def keys_changed(self) -> None:
        kc = self.keyboard_config
        if kc:
            kc.compute_modifier_map()
            kc.compute_modifier_keynames()
        log("keys_changed() updated keyboard config=%s", self.keyboard_config)

    def make_keymask_match(self, modifier_list, ignored_modifier_keycode=None, ignored_modifier_keynames=None):
        kc = self.keyboard_config
        if kc and kc.enabled:
            kc.make_keymask_match(modifier_list, ignored_modifier_keycode, ignored_modifier_keynames)

    def set_default_keymap(self):
        log("set_default_keymap() keyboard_config=%s", self.keyboard_config)
        kc = self.keyboard_config
        if kc:
            kc.set_default_keymap()
        return kc

    def is_modifier(self, keyname: str, keycode: int) -> bool:
        if keyname in DEFAULT_MODIFIER_MEANINGS:
            return True
        # keyboard config should always exist if we are here?
        kc = self.keyboard_config
        if kc:
            return kc.is_modifier(keycode)
        return False

    def set_keymap(self, current_keyboard_config, keys_pressed, force: bool = False,
                   translate_only: bool = False) -> None:
        kc = self.keyboard_config
        log("set_keymap%s keyboard_config=%s", (current_keyboard_config, keys_pressed, force, translate_only), kc)
        if kc and kc.enabled:
            current_id = ""
            if current_keyboard_config and current_keyboard_config.enabled:
                current_id = current_keyboard_config.get_hash()
            keymap_id = kc.get_hash()
            log("current keyboard id=%s, new keyboard id=%s", current_id, keymap_id)
            if force or not current_id or keymap_id != current_id:
                kc.keys_pressed = keys_pressed
                kc.set_keymap(translate_only)
                kc.owner = self.uuid
            else:
                log.info("keyboard mapping already configured (skipped)")
                self.keyboard_config = current_keyboard_config

    def get_keycode(self, client_keycode: int, keyname: str, pressed: bool,
                    modifiers, keyval, keystr: str, group: int) -> tuple[int, int]:
        kc = self.keyboard_config
        if kc is None:
            log.info("ignoring client key %s / %s since keyboard is not configured", client_keycode, keyname)
            return -1, 0
        return kc.get_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)

    def update_mouse(self, wid: int, x: int, y: int, rx: int, ry: int) -> None:
        log("update_mouse(%s, %i, %i, %i, %i) current=%s, client=%i",
            wid, x, y, rx, ry, self.mouse_last_position, self.counter)
        if self.mouse_last_position != (x, y) or self.mouse_last_relative_position != (rx, ry):
            self.mouse_last_position = (x, y)
            self.mouse_last_position = (rx, ry)
            self.send_async("pointer-position", wid, x, y, rx, ry)



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

def qrencode(s):
    try:
        from xpra.net.qrcode.qrencode import encode_image
    except ImportError:
        return None
    return encode_image(s)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from xpra.net.ssh.paramiko_client import keymd5
from xpra.net.ssh.util import get_default_keyfiles


class SSHTest(unittest.TestCase):

    def test_keymd5(self):
        k = AdHocStruct()
        k.get_fingerprint = lambda : b"abcd"
        assert keymd5(k).startswith("MD5:")

    def test_default_keyfiles(self):
        assert get_default_keyfiles()


def main():
    unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import logging
import traceback
from time import monotonic
from threading import Lock

from xpra.util.objects import typedict
from xpra.util.str_fn import csv, repr_ellipsized
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.log import Logger, set_global_logging_handler
from xpra.net.common import PacketType, LOG_PACKET_TYPE
from xpra.net.compression import Compressed

log = Logger("client")


class RemoteLogging(StubClientMixin):
    """
    Mixin for remote logging support,
    either sending local logging events to the server,
    or receiving logging events from the server.
    """
    PREFIX = "logging"

    def __init__(self):
        self.remote_logging = "no"
        self.in_remote_logging = False
        self.local_logging = None
        self.logging_lock = Lock()
        self.log_both = False
        self.request_server_log = False
        self.monotonic_start_time = monotonic()

    def init(self, opts) -> None:
        self.remote_logging = opts.remote_logging
        self.log_both = (opts.remote_logging or "").lower() == "both"

    def cleanup(self) -> None:
        ll = self.local_logging
        log("cleanup() local_logging=%s", ll)
        if ll:
            self.local_logging = None
            set_global_logging_handler(ll)

    def parse_server_capabilities(self, c: typedict) -> bool:
        receive = c.boolget("remote-logging.receive")
        send = c.boolget("remote-logging.send")
        v = c.get("remote-logging")
        if isinstance(v, dict):
            c = typedict(v)
            receive = c.boolget("receive")
            send = c.boolget("send")
        if self.remote_logging.lower() in ("send", "both", "yes", "true", "on") and receive:
            # check for debug:
            from xpra.log import is_debug_enabled
            conflict = tuple(v for v in ("network", "crypto", "websocket", "quic") if is_debug_enabled(v))
            if conflict:
                log.warn("Warning: cannot enable remote logging")
                log.warn(" because debug logging is enabled for: %s", csv(conflict))
                return True
            if LOG_PACKET_TYPE:
                log.warn("Warning: cannot enable remote logging")
                log.warn(f" because {LOG_PACKET_TYPE=}")
                return True
            log.info("enabled remote logging")
            if not self.log_both:
                log.info(" see server log file for further output")
            self.local_logging = set_global_logging_handler(self.remote_logging_handler)
        elif self.remote_logging.lower() == "receive":
            self.request_server_log = send
            if not self.request_server_log:
                log.warn("Warning: cannot receive log output from the server")
                log.warn(" the feature is not enabled or not supported by the server")
            else:
                self.after_handshake(self.start_receiving_logging)  # pylint: disable=no-member
        return True

    def start_receiving_logging(self) -> None:
        self.add_packets("logging-event")
        self.add_legacy_alias("logging", "logging-event")
        self.send("logging-control", "start")

    def _process_logging_event(self, packet: PacketType) -> None:
        assert not self.local_logging, "cannot receive logging packets when forwarding logging!"
        level = int(packet[1])
        msg = packet[2]
        prefix = "server: "
        if len(packet) >= 4:
            dtime = int(packet[3])
            prefix += "@%02i.%03i " % ((dtime // 1000) % 60, dtime % 1000)
        try:
            if isinstance(msg, (tuple, list)):
                dmsg = " ".join(str(x) for x in msg)
            else:
                dmsg = str(msg)
            for line in dmsg.splitlines():
                self.do_log(level, prefix + line)
        except Exception as e:
            log("log message decoding error", exc_info=True)
            log.error("Error: failed to parse logging message:")
            log.error(" %s", repr_ellipsized(msg))
            log.estr(e)

    def do_log(self, level: int, line) -> None:
        with self.logging_lock:
            log.log(level, line)

    def remote_logging_handler(self, logger_log, level: int, msg: str, *args, **kwargs) -> None:
        # prevent loops (if our send call ends up firing another logging call):
        if self.in_remote_logging:
            return
        self.in_remote_logging = True
        ll = self.local_logging

        def local_warn(*warn_args) -> None:
            if ll:
                ll(logger_log, logging.WARNING, *warn_args)

        try:
            if not kwargs.pop("remote", True):
                # keyword is telling us not to forward it!
                if ll:
                    try:
                        ll(logger_log, level, msg, *args, **kwargs)
                    except Exception as e:
                        local_warn("Warning: failed to log message locally")
                        local_warn(" %s", e)
                        local_warn(" %s", msg)
                return

            dtime = int(1000 * (monotonic() - self.monotonic_start_time))
            data: str | Compressed
            if args:
                data = msg % args
            else:
                data = msg
            if len(data) >= 32:
                try:
                    data = self.compressed_wrapper("text", data.encode("utf8"), level=1)
                except Exception:
                    pass
            self.send("logging", level, data, dtime)
            exc_info = kwargs.get("exc_info")
            if exc_info is True:
                exc_info = sys.exc_info()
            if exc_info and exc_info[0]:
                for x in traceback.format_tb(exc_info[2]):
                    self.send("logging", level, x, dtime)
                try:
                    etypeinfo = exc_info[0].__name__
                except AttributeError:
                    etypeinfo = str(exc_info[0])
                self.send("logging", level, f"{etypeinfo}: {exc_info[1]}", dtime)
            if self.log_both and ll:
                ll(logger_log, level, msg, *args, **kwargs)
        except Exception as e:
            if self.exit_code is not None:
                # errors can happen during exit, don't care
                return
            local_warn("Warning: failed to send logging packet:")
            local_warn(f" {e}")
            local_warn(f" original unformatted message: {msg}")
            if args:
                local_warn(f" {len(args)} arguments: {args}")
            else:
                local_warn(" (no arguments)")
            if ll:
                try:
                    ll(logger_log, level, msg, *args, **kwargs)
                except Exception:
                    pass
            try:
                exc_info = sys.exc_info()
                for x in traceback.format_tb(exc_info[2]):
                    for v in x.splitlines():
                        local_warn(v)
            except Exception:
                pass
        finally:
            self.in_remote_logging = False



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from pyglet.window import Window, event
from pyglet.image import ImageData, SolidColorImagePattern
from pyglet.sprite import Sprite

from xpra.client.pyglet.keys import keynames
from xpra.log import Logger

log = Logger()


class ClientWindow(Window):
    def __init__(self, client, wid: int, x: int, y: int, w: int, h: int, metadata: dict):
        super().__init__(visible=False)
        self.client = client
        self.send = client.send
        self.wid = wid
        self.metadata = metadata
        self.image = SolidColorImagePattern((128, 0, 0, 128)).create_image(w, h)
        self.sprite = Sprite(img=self.image, x=0, y=0)
        #self.image_data = self.image.get_image_data()
        #self.texture = self.image.get_texture()
        # window.config.alpha_size = 8
        # if metadata.get("override-redirect"):
        #    how?
        self.set_caption(metadata.get("title", ""))
        self.set_location(x, y)
        self.set_size(w, h)

        # A logger class, which prints events to stdout or to a file:
        win_event_logger = event.WindowEventLogger()
        self.push_handlers(win_event_logger)

    def show(self) -> None:
        self.set_visible(True)

    def on_close(self) -> None:
        self.client.send("close-window", self.wid)

    def on_activate(self) -> None:
        self.client.update_focus(self.wid)

    def on_show(self) -> None:
        x, y = self.get_location()
        w, h = self.get_size()
        self.client.send("map-window", self.wid, x, y, w, h, {}, ())

    def on_hide(self) -> None:
        self.client.send("unmap-window", self.wid)

    def on_key_press(self, keysym: int, modifiers: int) -> None:
        self.on_key_event(True, keysym, modifiers)

    def on_key_release(self, keysym: int, modifiers: int) -> None:
        self.on_key_event(False, keysym, modifiers)

    def on_key_event(self, pressed: bool, keysym: int, modifiers: int) -> None:
        keyname = keynames.get(keysym)
        if not keyname:
            log.info(f"unknown key {keysym}")
            return
        mods = ()
        keyval = 0
        string = ""
        keycode = 0
        group = 0
        self.client.send("key-action", self.wid, keyname, pressed, mods, keyval, string, keycode, group)

    def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> None:
        device_id = -1
        seq = 0
        pos = (x, y, dx, dy)
        self.client.send("pointer", device_id, seq, self.wid, pos, {})

    def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> None:
        self.on_mouse_button_event(True, x, y, button, modifiers)

    def on_mouse_release(self, x: int, y: int, button: int, modifiers: int) -> None:
        self.on_mouse_button_event(False, x, y, button, modifiers)

    def on_mouse_button_event(self, pressed: bool, x: int, y: int, button: int, modifiers: int) -> None:
        seq = 0
        pos = (x, y)
        self.client.send("pointer-button", -1, seq, self.wid, button, pressed, pos, {})

    def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float) -> None:
        button = 4
        pointer = (x, y)
        modifiers = ()
        buttons = ()
        for state in True, False:
            self.send_button(-1, self.wid, button, state, pointer, modifiers, buttons, {})

    def on_move(self, x: int, y: int) -> None:
        counter = 0
        props = {}
        w, h = self.get_size()
        state = ()
        skip_geometry = False
        self.client.send("configure-window", self.wid, x, y, w, h, props, counter, state, skip_geometry)

    def on_resize(self, w: int, h: int) -> None:
        counter = 0
        props = {}
        x, y = self.get_location()
        state = ()
        skip_geometry = False
        self.client.send("configure-window", self.wid, x, y, w, h, props, counter, state, skip_geometry)

    def draw(self, x, y, w, h, coding, data, stride) -> None:
        if coding in ("png", "jpg", "webp"):
            from PIL import Image
            from io import BytesIO
            img = Image.open(BytesIO(data))
            if img.mode == "RGBA":
                coding = "rgb32"
                stride = w * 4
            else:
                assert img.mode == "RGB"
                coding = "rgb24"
                stride = w * 3
            flipped = img.transpose(Image.Transpose.FLIP_TOP_BOTTOM)
            data = flipped.tobytes("raw", img.mode)
        if coding == "rgb24":
            image = ImageData(w, h, "RGB", data, pitch=stride or w*3)
        elif coding == "rgb32":
            image = ImageData(w, h, "RGBA", data, pitch=stride or w*4)
        else:
            raise ValueError(f"invalid encoding {coding!r}")
        bh = self.get_size()[1]
        self.sprite = Sprite(image, x=x, y=bh - y - h)
        self.sprite.draw()
        self.flip()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.os_util import gi_import
from xpra.util.io import get_util_logger
from xpra.gtk.signals import register_os_signals
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.platform.gui import force_focus

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")
GdkPixbuf = gi_import("GdkPixbuf")

log = get_util_logger()


class PasswordInputDialogWindow(Gtk.Dialog):

    def __init__(self, title="Title", prompt="", icon=""):
        super().__init__()
        self.set_border_width(20)
        self.set_position(Gtk.WindowPosition.CENTER)
        self.connect("delete-event", self.quit)
        self.set_default_size(400, 150)
        self.set_title(title)
        add_close_accel(self, self.cancel)
        self.password = ""

        icon_pixbuf = get_icon_pixbuf(icon)
        if icon_pixbuf:
            self.set_icon(icon_pixbuf)

        vbox = self.get_content_area()
        vbox.set_spacing(10)

        def al(text: str, font="sans 14", xalign=0) -> None:
            lbl = label(text, font=font)
            align = Gtk.Alignment(xalign=xalign, yalign=0.5, xscale=0.0, yscale=0)
            align.add(lbl)
            vbox.add(align)
            align.show_all()

        # window title is visible so this would be redundant:
        # al(title, "sans 18", 0.5)
        al(prompt, "sans 14")
        self.password_input = Gtk.Entry()
        self.password_input.set_max_length(255)
        self.password_input.set_width_chars(32)
        self.password_input.connect('activate', self.password_activate)
        self.password_input.connect('changed', self.password_changed)
        self.password_input.set_visibility(False)
        vbox.add(self.password_input)

        self.confirm_btn = self.add_button("Confirm", 0)
        self.set_default(self.confirm_btn)
        self.set_focus(self.confirm_btn)
        self.cancel_btn = self.add_button("Cancel", 1)

    def show(self):
        log("PasswordInputDialogWindow.show()")
        self.show_all()

        def show():
            force_focus()
            self.present()
            self.password_input.grab_focus()

        GLib.idle_add(show)

    def run(self) -> str:
        r = super().run()
        if r == 0:
            return self.get_password()
        return ""

    def quit(self, *args) -> bool:
        log("quit%s", args)
        self.destroy()
        return True

    def signal_quit(self, *args) -> None:
        log("quit%s", args)
        self.close()

    def cancel(self, *args):
        log("cancel%s", args)
        self.cancel_btn.activate()

    def password_activate(self, *_args):
        self.confirm_btn.activate()

    def password_changed(self, *_args):
        self.password = self.password_input.get_text()

    def get_password(self) -> str:
        return self.password


def show_pass_dialog(argv):
    from xpra.platform.gui import ready as gui_ready, init as gui_init, set_default_icon

    set_default_icon("authentication.png")
    gui_init()

    log("show_pass_dialog(%s)", argv)

    def arg(n):
        if len(argv) <= n:
            return ""
        return argv[n].replace("\\n\\r", "\\n").replace("\\n", "\n")

    title = arg(0) or "Enter Password"
    prompt = arg(1)
    icon = arg(2)
    app = PasswordInputDialogWindow(title, prompt, icon)
    register_os_signals(app.signal_quit, "Password Dialog")
    gui_ready()
    app.show()
    p = app.run()
    if p is not None:
        sys.stdout.write(p)
        sys.stdout.flush()
        return 0
    return 1


def main():
    from xpra.platform import program_context
    with program_context("Password-Input-Dialog", "Password Input Dialog"):
        if "-v" in sys.argv:
            from xpra.log import enable_debug_for
            enable_debug_for("util")

        try:
            return show_pass_dialog(sys.argv[1:])
        except KeyboardInterrupt:
            return 1


if __name__ == "__main__":
    v = main()
    sys.exit(v)



#!/usr/bin/env python3
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.platform import program_context
from xpra.platform.gui import force_focus
from xpra.gtk.util import GRAB_STATUS_STRING
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import label
from xpra.gtk.pixbuf import get_icon_pixbuf

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")


def make_grab_window():
    window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
    window.set_size_request(600, 200)
    window.set_title("Grabs")
    window.connect("delete_event", Gtk.main_quit)
    window.add_events(Gdk.EventMask.ALL_EVENTS_MASK)
    window.set_position(Gtk.WindowPosition.CENTER)
    icon = get_icon_pixbuf("pointer.png")
    if icon:
        window.set_icon(icon)
    vbox = Gtk.VBox(homogeneous=False, spacing=0)
    hbox = Gtk.HBox(homogeneous=False, spacing=0)
    vbox.pack_start(hbox, expand=False, fill=False, padding=10)

    def keyevent_info(event):
        keyval = event.keyval
        keycode = event.hardware_keycode
        keyname = Gdk.keyval_name(keyval)
        return "%i:%s" % (keycode, keyname)

    def key_pressed(_window, event):
        event_label.set_text("key_pressed: %s" % keyevent_info(event))

    window.connect("key-press-event", key_pressed)

    def key_released(_window, event):
        event_label.set_text("key_released: %s" % keyevent_info(event))

    window.connect("key-press-event", key_pressed)
    window.connect("key-release-event", key_released)

    def motion_notify(_window, event):
        event_label.set_text("motion: %i,%i" % (event.x_root, event.y_root))

    window.connect("motion-notify-event", motion_notify)

    grab_pointer_btn = Gtk.Button(label="grab pointer")

    def grab_pointer(*args):
        action_label.set_text("grab_pointer%s" % str(args))

        def do_grab():
            event_mask = Gdk.EventMask.BUTTON_PRESS_MASK | Gdk.EventMask.BUTTON_RELEASE_MASK
            v = Gdk.pointer_grab(window.get_window(), False, event_mask, None, None, 0)
            # Gdk.BUTTON_PRESS_MASK | Gdk.BUTTON_RELEASE_MASK | Gdk.KEY_PRESS_MASK \
            # Gdk.KEY_RELEASE_MASK | Gdk.ENTER_NOTIFY_MASK)
            # | Gdk.ENTER_NOTIFY_MASK
            # Gdk.ALL_EVENTS_MASK
            action_label.set_text("pointer_grab() returned %s" % GRAB_STATUS_STRING.get(v, v))
            GLib.timeout_add(10 * 1000, Gdk.pointer_ungrab, 0)

        print("will grab in 5 seconds!")
        GLib.timeout_add(5 * 1000, do_grab)

    grab_pointer_btn.connect('clicked', grab_pointer)
    hbox.pack_start(grab_pointer_btn, expand=False, fill=False, padding=10)

    ungrab_pointer_btn = Gtk.Button(label="ungrab pointer")

    def ungrab_pointer(*_args):
        v = Gdk.pointer_ungrab(0)
        action_label.set_text("pointer_ungrab(0)=%s" % GRAB_STATUS_STRING.get(v, v))
        window.unmaximize()

    ungrab_pointer_btn.connect('clicked', ungrab_pointer)
    hbox.pack_start(ungrab_pointer_btn, expand=False, fill=False, padding=10)

    grab_keyboard_btn = Gtk.Button(label="grab keyboard")

    def grab_keyboard(*_args):
        v = Gdk.keyboard_grab(window.get_window(), True, 0)
        action_label.set_text("keyboard_grab(..)=%s" % GRAB_STATUS_STRING.get(v, v))
        GLib.timeout_add(10 * 1000, Gdk.keyboard_ungrab, 0)

    grab_keyboard_btn.connect('clicked', grab_keyboard)
    hbox.pack_start(grab_keyboard_btn, expand=False, fill=False, padding=10)

    ungrab_keyboard_btn = Gtk.Button(label="ungrab keyboard")

    def ungrab_keyboard(*_args):
        v = Gdk.keyboard_ungrab(0)
        action_label.set_text("keyboard_ungrab(0)=%s" % GRAB_STATUS_STRING.get(v, v))

    ungrab_keyboard_btn.connect('clicked', ungrab_keyboard)
    hbox.pack_start(ungrab_keyboard_btn, expand=False, fill=False, padding=10)

    vbox.add(label("Last action:"))
    action_label = label("")
    vbox.add(action_label)

    vbox.add(label("Last event:"))
    event_label = label("")
    vbox.add(event_label)

    window.add(vbox)
    return window


def main():
    with program_context("grabs", "Grabs"):
        w = make_grab_window()

        def show_with_focus():
            force_focus()
            w.show_all()
            w.present()

        add_close_accel(w, Gtk.main_quit)
        from xpra.gtk.signals import quit_on_signals
        quit_on_signals("grabs test window")
        GLib.idle_add(show_with_focus)
        Gtk.main()
        return 0


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable-msg=E1101

import os.path
from time import monotonic

from xpra.util.parsing import parse_scaling_value, from0to100
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.common import ConnectionMessage, noop
from xpra.util.io import load_binary_file
from xpra.net.common import PacketType
from xpra.util.stats import std_unit
from xpra.scripts.config import str_to_bool, FALSE_OPTIONS, TRUE_OPTIONS
from xpra.net.control.common import ArgsControlCommand, ControlError
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

log = Logger("command")

TOGGLE_FEATURES = (
    "bell", "randr", "cursors", "notifications", "clipboard",
    "start-new-commands", "client-shutdown", "webcam",
)


class ServerBaseControlCommands(StubServerMixin):
    """
    Control commands for ServerBase
    """
    PREFIX = "control"

    def setup(self) -> None:
        self.add_control_commands()

    def add_control_commands(self) -> None:
        def parse_boolean_value(v):
            if str(v).lower() in TRUE_OPTIONS:
                return True
            if str(v).lower() in FALSE_OPTIONS:
                return False
            raise ControlError(f"a boolean is required, not {v!r}")

        def parse_4intlist(v) -> list:
            if not v:
                return []
            intlist = []
            # ie: v = " (0,10,100,20), (200,300,20,20)"
            while v:
                v = v.strip().strip(",").strip()  # ie: "(0,10,100,20)"
                lp = v.find("(")
                assert lp == 0, "invalid leading characters: %s" % v[:lp]
                rp = v.find(")")
                assert (lp + 1) < rp
                item = v[lp + 1:rp].strip()  # "0,10,100,20"
                items = [int(x) for x in item]  # 0,10,100,20
                assert len(items) == 4, f"expected 4 numbers but got {len(items)}"
                intlist.append(items)
            return intlist

        for cmd in (
                ArgsControlCommand("focus", "give focus to the window id", validation=[int]),
                ArgsControlCommand("map", "maps the window id", validation=[int]),
                ArgsControlCommand("unmap", "unmaps the window id", validation=[int]),
                # window source:
                ArgsControlCommand("suspend", "suspend screen updates", max_args=0),
                ArgsControlCommand("resume", "resume screen updates", max_args=0),
                ArgsControlCommand("ungrab", "cancels any grabs", max_args=0),
                # server globals:
                ArgsControlCommand("readonly", "set readonly state for client(s)", min_args=1, max_args=1,
                                   validation=[parse_boolean_value]),
                ArgsControlCommand("idle-timeout", "set the idle timeout", validation=[int]),
                ArgsControlCommand("server-idle-timeout", "set the server idle timeout", validation=[int]),
                ArgsControlCommand("start-env", "modify the environment used to start new commands", min_args=2),
                ArgsControlCommand("start", "executes the command arguments in the server context", min_args=1),
                ArgsControlCommand("start-child",
                                   "executes the command arguments in the server context, as a 'child' (honouring exit-with-children)",
                                   min_args=1),
                ArgsControlCommand("toggle-feature",
                                   "toggle a server feature on or off, one of: %s" % csv(TOGGLE_FEATURES), min_args=1,
                                   max_args=2, validation=[str, parse_boolean_value]),
                # network and transfers:
                ArgsControlCommand("print", "sends the file to the client(s) for printing", min_args=1),
                ArgsControlCommand("open-url", "open the URL on the client(s)", min_args=1, max_args=2),
                ArgsControlCommand("send-file", "sends the file to the client(s)", min_args=1, max_args=4),
                ArgsControlCommand("send-notification", "sends a notification to the client(s)", min_args=4, max_args=5,
                                   validation=[int]),
                ArgsControlCommand("close-notification",
                                   "send the request to close an existing notification to the client(s)", min_args=1,
                                   max_args=2, validation=[int]),
                ArgsControlCommand("compression", "sets the packet compressor", min_args=1, max_args=1),
                ArgsControlCommand("encoder", "sets the packet encoder", min_args=1, max_args=1),
                ArgsControlCommand("clipboard-direction", "restrict clipboard transfers", min_args=1, max_args=1),
                ArgsControlCommand("clipboard-limits", "restrict clipboard transfers size", min_args=2, max_args=2,
                                   validation=[int, int]),
                ArgsControlCommand("set-lock", "modify the lock attribute", min_args=1, max_args=1),
                ArgsControlCommand("set-sharing", "modify the sharing attribute", min_args=1, max_args=1),
                ArgsControlCommand("set-ui-driver", "set the client connection driving the session", min_args=1,
                                   max_args=1),
                # session and clients:
                ArgsControlCommand("client", "forwards a control command to the client(s)", min_args=1),
                ArgsControlCommand("client-property", "set a client property", min_args=4, max_args=5,
                                   validation=[int]),
                ArgsControlCommand("name", "set the session name", min_args=1, max_args=1),
                ArgsControlCommand("key", "press or unpress a key", min_args=1, max_args=2),
                ArgsControlCommand("audio-output", "control audio forwarding", min_args=1, max_args=2),
                # windows:
                ArgsControlCommand("workspace", "move a window to a different workspace", min_args=2, max_args=2,
                                   validation=[int, int]),
                ArgsControlCommand("close", "close a window", min_args=1, max_args=1, validation=[int]),
                ArgsControlCommand("delete", "delete a window", min_args=1, max_args=1, validation=[int]),
                ArgsControlCommand("move", "move a window", min_args=3, max_args=3, validation=[int, int, int]),
                ArgsControlCommand("resize", "resize a window", min_args=3, max_args=3, validation=[int, int, int]),
                ArgsControlCommand("moveresize", "move and resize a window", min_args=5, max_args=5,
                                   validation=[int, int, int, int, int]),
                ArgsControlCommand("scaling-control", "set the scaling-control aggressiveness (from 0 to 100)",
                                   min_args=1, validation=[from0to100]),
                ArgsControlCommand("scaling", "set a specific scaling value", min_args=1,
                                   validation=[parse_scaling_value]),
                ArgsControlCommand("auto-refresh", "set a specific auto-refresh value", min_args=1, validation=[float]),
                ArgsControlCommand("refresh", "refresh some or all windows", min_args=0),
                ArgsControlCommand("encoding", "picture encoding", min_args=2),
                ArgsControlCommand("request-update", "request a screen update using a specific encoding", min_args=3),
                ArgsControlCommand("video-region-enabled", "enable video region", min_args=2, max_args=2,
                                   validation=[int, parse_boolean_value]),
                ArgsControlCommand("video-region-detection", "enable video detection", min_args=2, max_args=2,
                                   validation=[int, parse_boolean_value]),
                ArgsControlCommand("video-region-exclusion-zones",
                                   "set window regions to exclude from video regions: 'WID,(x,y,w,h),(x,y,w,h),..', ie: '1 (0,10,100,20),(200,300,20,20)'",
                                   min_args=2, max_args=2, validation=[int, parse_4intlist]),
                ArgsControlCommand("video-region", "set the video region", min_args=5, max_args=5,
                                   validation=[int, int, int, int, int]),
                ArgsControlCommand("reset-video-region", "reset video region heuristics", min_args=1, max_args=1,
                                   validation=[int]),
                ArgsControlCommand("lock-batch-delay", "set a specific batch delay for a window", min_args=2,
                                   max_args=2, validation=[int, int]),
                ArgsControlCommand("unlock-batch-delay",
                                   "let the heuristics calculate the batch delay again for a window (following a 'lock-batch-delay')",
                                   min_args=1, max_args=1, validation=[int]),
                ArgsControlCommand("remove-window-filters", "remove all window filters", min_args=0, max_args=0),
                ArgsControlCommand("add-window-filter", "add a window filter", min_args=4, max_args=5),
        ):
            cmd.do_run = getattr(self, "control_command_%s" % cmd.name.replace("-", "_"), noop)
            if cmd.do_run != noop:
                self.add_control_command(cmd.name, cmd)
        # encoding bits:
        for name in (
                "quality", "min-quality", "max-quality",
                "speed", "min-speed", "max-speed",
        ):
            fn = getattr(self, "control_command_%s" % name.replace("-", "_"), noop)
            if fn != noop:
                self.add_control_command(name, ArgsControlCommand(name, "set encoding %s (from 0 to 100)" % name, run=fn,
                                                                  min_args=1, validation=[from0to100]))

    #########################################
    # Control Commands
    #########################################
    def control_command_focus(self, wid: int) -> str:
        if self.readonly:
            return "focus request denied by readonly mode"
        if not isinstance(wid, int):
            raise ValueError(f"argument should have been an int, but found {type(wid)}")
        self._focus(None, wid, None)
        return f"gave focus to window {wid}"

    def control_command_map(self, wid: int) -> str:
        if self.readonly:
            return "map request denied by readonly mode"
        if not isinstance(wid, int):
            raise ValueError(f"argument should have been an int, but found {type(wid)}")
        window = self._id_to_window.get(wid)
        assert window, f"window {wid} not found"
        if window.is_tray():
            return f"cannot map tray window {wid}"
        if window.is_OR():
            return f"cannot map override redirect window {wid}"
        window.show()
        # window.set_owner(dm)
        # iconic = window.get_property("iconic")
        # if iconic:
        #    window.set_property("iconic", False)
        # w, h = window.get_geometry()[2:4]
        # self.refresh_window_area(window, 0, 0, w, h)
        self.repaint_root_overlay()
        return "mapped window %s" % wid

    def control_command_unmap(self, wid: int) -> str:
        if self.readonly:
            return "unmap request denied by readonly mode"
        if not isinstance(wid, int):
            raise ValueError(f"argument should have been an int, but found {type(wid)}")
        window = self._id_to_window.get(wid)
        assert window, f"window {wid} not found"
        if window.is_tray():
            return f"cannot unmap tray window {wid}"
        if window.is_OR():
            return f"cannot unmap override redirect window {wid}"
        window.hide()
        self.repaint_root_overlay()
        return f"unmapped window {wid}"

    def control_command_suspend(self) -> str:
        for csource in tuple(self._server_sources.values()):
            csource.suspend(True, self._id_to_window)
        count = len(self._server_sources)
        return f"suspended {count} clients"

    def control_command_resume(self) -> str:
        for csource in tuple(self._server_sources.values()):
            csource.resume(True, self._id_to_window)
        count = len(self._server_sources)
        return f"resumed {count} clients"

    def control_command_ungrab(self) -> str:
        for csource in tuple(self._server_sources.values()):
            csource.pointer_ungrab(-1)
        count = len(self._server_sources)
        return f"ungrabbed {count} clients"

    def control_command_readonly(self, onoff) -> str:
        log("control_command_readonly(%s)", onoff)
        self.readonly = onoff
        msg = f"server readonly: {onoff}"
        log.info(msg)
        return msg

    def control_command_idle_timeout(self, t: int) -> str:
        self.idle_timeout = t
        for csource in tuple(self._server_sources.values()):
            csource.idle_timeout = t
            csource.schedule_idle_timeout()
        return f"idle-timeout set to {t}"

    def control_command_server_idle_timeout(self, t: int) -> str:
        self.server_idle_timeout = t
        reschedule = len(self._server_sources) == 0
        self.reset_server_timeout(reschedule)
        return f"server-idle-timeout set to {t}"

    def control_command_start_env(self, action: str = "set", var_name: str = "", value=None) -> str:
        assert var_name, "the environment variable name must be specified"
        if action == "unset":
            assert value is None, f"invalid number of arguments for {action}"
            if self.start_env.pop(var_name, None) is None:
                return f"{var_name!r} is not set"
            return f"{var_name} unset"
        if action == "set":
            assert value, "the value must be specified"
            self.start_env[var_name] = value
            return f"{var_name}={value}"
        return f"invalid start-env subcommand {action!r}"

    def control_command_start(self, *args) -> str:
        return self.do_control_command_start(True, *args)

    def control_command_start_child(self, *args) -> str:
        return self.do_control_command_start(False, *args)

    def do_control_command_start(self, ignore, *args) -> str:
        if not self.start_new_commands:
            raise ControlError("this feature is currently disabled")
        proc = self.start_command(" ".join(args), args, ignore)
        if not proc:
            raise ControlError("failed to start new child command " + str(args))
        return "new %scommand started with pid=%s" % (["child ", ""][ignore], proc.pid)

    def control_command_toggle_feature(self, feature, state=None) -> str:
        log("control_command_toggle_feature(%s, %s)", feature, state)
        if feature not in TOGGLE_FEATURES:
            msg = f"invalid feature {feature!r}"
            log.warn(msg)
            return msg
        fn = feature.replace("-", "_")
        if not hasattr(self, feature):
            msg = f"attribute {feature!r} not found - bug?"
            log.warn(msg)
            return msg
        cur = getattr(self, fn, None)
        if state is None:
            # if the new state is not specified, just negate the value
            state = not cur
        setattr(self, fn, state)
        self.setting_changed(feature, state)
        return f"{feature} set to {state}"

    def _control_get_sources(self, client_uuids_str, _attr=None):
        # find the client uuid specified as a string:
        if client_uuids_str == "UI":
            sources = [ss for ss in self._server_sources.values() if ss.ui_client]
        elif client_uuids_str == "*":
            sources = self._server_sources.values()
        else:
            client_uuids = client_uuids_str.split(",")
            sources = [ss for ss in self._server_sources.values() if ss.uuid in client_uuids]
            uuids = tuple(ss.uuid for ss in sources)
            notfound = any(x for x in client_uuids if x not in uuids)
            if notfound:
                log.warn(f"Warning: client connection not found for uuid(s): {notfound}")
        return sources

    def control_command_send_notification(self, nid: int, title: str, message: str, client_uuids) -> str:
        if not self.notifications:
            msg = "notifications are disabled"
            log(msg)
            return msg
        sources = self._control_get_sources(client_uuids)
        log("control_command_send_notification(%i, %s, %s, %s) will send to sources %s (matching %s)",
            nid, title, message, client_uuids, sources, client_uuids)
        count = 0
        for source in sources:
            if source.notify(0, nid, "control channel", 0, "", title, message, [], {}, 10, ""):
                count += 1
        msg = f"notification id {nid}: message sent to {count} clients"
        log(msg)
        return msg

    def control_command_close_notification(self, nid: int, client_uuids) -> str:
        if not self.notifications:
            msg = "notifications are disabled"
            log(msg)
            return msg
        sources = self._control_get_sources(client_uuids)
        log("control_command_close_notification(%s, %s) will send to %s", nid, client_uuids, sources)
        for source in sources:
            source.notify_close(nid)
        msg = f"notification id {nid}: close request sent to {len(sources)} clients"
        log(msg)
        return msg

    def control_command_open_url(self, url: str, client_uuids="*") -> str:
        # find the clients:
        sources = self._control_get_sources(client_uuids)
        if not sources:
            raise ControlError(f"no clients found matching: {client_uuids!r}")
        clients = 0
        for ss in sources:
            if hasattr(ss, "send_open_url") and ss.send_open_url(url):
                clients += 1
        return f"url sent to {clients} clients"

    def control_command_send_file(self, filename: str, openit="open", client_uuids="*", maxbitrate=0) -> str:
        # we always get the values as strings from the command interface,
        # but those may actually be utf8 encoded binary strings,
        # so we may have to do an ugly roundtrip:
        openit = str(openit).lower() in ("open", "true", "1")
        return self.do_control_file_command("send file", client_uuids, filename, "file_transfer", (False, openit))

    def control_command_print(self, filename: str, printer="", client_uuids="*",
                              maxbitrate=0, title="", *options_strs) -> str:
        # FIXME: printer and bitrate are ignored
        # parse options into a dict:
        options = {}
        for arg in options_strs:
            argp = arg.split("=", 1)
            if len(argp) == 2 and len(argp[0]) > 0:
                options[argp[0]] = argp[1]
        return self.do_control_file_command("print", client_uuids, filename, "printing", (True, True, options))

    def do_control_file_command(self, command_type, client_uuids, filename, source_flag_name, send_file_args) -> str:
        # find the clients:
        sources = self._control_get_sources(client_uuids)
        if not sources:
            raise ControlError(f"no clients found matching: {client_uuids!r}")

        filelog = Logger("command", "file")

        def checksize(file_size):
            if file_size > self.file_transfer.file_size_limit:
                raise ControlError("file '%s' is too large: %sB (limit is %sB)" % (
                    filename, std_unit(file_size), std_unit(self.file_transfer.file_size_limit)))

        # find the file and load it:
        actual_filename = os.path.abspath(os.path.expanduser(filename))
        try:
            stat = os.stat(actual_filename)
            filelog("os.stat(%s)=%s", actual_filename, stat)
        except os.error:
            filelog("os.stat(%s)", actual_filename, exc_info=True)
        else:
            checksize(stat.st_size)
        if not os.path.exists(actual_filename):
            raise ControlError(f"file {filename!r} does not exist")
        data = load_binary_file(actual_filename)
        if not data:
            raise ControlError(f"no data loaded from {actual_filename!r}")
        # verify size:
        file_size = len(data)
        checksize(file_size)
        # send it to each client:
        for ss in sources:
            # ie: ServerSource.file_transfer (found in FileTransferAttributes)
            #     and ServerSource.remote_file_transfer (found in FileTransferHandler)
            server_support = getattr(ss, source_flag_name, False)
            client_support = getattr(ss, f"remote_{source_flag_name}", False)
            if not (server_support and client_support):
                # skip the warning if the client is not interactive
                # (for now just check for 'top' client):
                if not hasattr(ss, source_flag_name) or ss.client_type == "top":
                    log_fn = filelog.debug
                else:
                    log_fn = filelog.warn
                log_fn(f"Warning: cannot {command_type} {filename!r} to {ss.client_type} client")
                log_fn(f" feature flag {source_flag_name!r}")
                if not server_support:
                    log_fn(" this feature is not supported by the server connection")
                if not client_support:
                    log_fn(f" client {ss.uuid} does not support this feature")
            elif file_size > ss.file_size_limit:
                filelog.warn(f"Warning: cannot {command_type} {filename!r}")
                filelog.warn(" client %s file size limit is %sB (file is %sB)",
                             ss, std_unit(ss.file_size_limit), std_unit(file_size))
            else:
                filelog(f"sending {filename} to {ss}")
                ss.send_file(filename, "", data, file_size, *send_file_args)
        return f"{command_type} of {filename!r} to {client_uuids} initiated"

    def control_command_remove_window_filters(self) -> str:
        # modify the existing list object,
        # which is referenced by all the sources
        count = len(self.window_filters)
        self.window_filters[:] = []
        return f"removed {count} window-filters"

    def control_command_add_window_filter(self, object_name: str, property_name: str, operator: str, value,
                                          client_uuids="") -> str:
        from xpra.server.window import filters  # pylint: disable=import-outside-toplevel
        window_filter = filters.get_window_filter(object_name, property_name, operator, value)
        # log("%s%s=%s", filters.get_window_filter, (object_name, property_name, operator, value), window_filter)
        if client_uuids == "*":
            # applies to all sources:
            self.window_filters.append(("*", window_filter))
        else:
            for client_uuid in client_uuids.split(","):
                self.window_filters.append((client_uuid, window_filter))
        return f"added window-filter: {window_filter} for client uuids={client_uuids}"

    def control_command_compression(self, compress: str) -> str:
        c = compress.lower()
        from xpra.net import compression  # pylint: disable=import-outside-toplevel
        opts = compression.get_enabled_compressors()  # ie: [lz4, zlib]
        if c not in opts:
            raise ControlError("compressor argument must be one of: " + csv(opts))
        for cproto in tuple(self._server_sources.keys()):
            cproto.enable_compressor(c)
        self.all_send_client_command(f"enable_{c}")
        return f"compressors set to {compression}"

    def control_command_encoder(self, encoder: str) -> str:
        e = encoder.lower()
        from xpra.net import packet_encoding  # pylint: disable=import-outside-toplevel
        opts = packet_encoding.get_enabled_encoders()  # ie: [rencodeplus, ]
        if e not in opts:
            raise ControlError("encoder argument must be one of: " + csv(opts))
        for cproto in tuple(self._server_sources.keys()):
            cproto.enable_encoder(e)
        self.all_send_client_command(f"enable_{e}")
        return f"encoders set to {encoder}"

    def all_send_client_command(self, *client_command) -> None:
        """ forwards the command to all clients """
        for source in tuple(self._server_sources.values()):
            # forwards to *the* client, if there is *one*
            if client_command[0] not in source.client_control_commands:
                log.info(f"client command {client_command!r} not forwarded to client {source} (not supported)")
            else:
                source.send_client_command(*client_command)

    def control_command_client(self, *args) -> str:
        self.all_send_client_command(*args)
        return f"client control command {args} forwarded to clients"

    def control_command_client_property(self, wid: int, uuid, prop: str, value, conv=None) -> str:
        wid = int(wid)
        conv_fn = {
            "int": int,
            "float": float,
            "": str,
        }.get(conv)
        assert conv_fn
        typeinfo = "%s " % (conv or "string")
        value = conv_fn(value)
        self.client_properties.setdefault(wid, {}).setdefault(uuid, {})[prop] = value
        return f"property {prop!r} set to {typeinfo} value {value!r} for window {wid}, client {uuid}"

    def control_command_name(self, name: str) -> str:
        self.session_name = name
        log.info(f"changed session name: {self.session_name!r}")
        # self.all_send_client_command("name", name)    not supported by any clients, don't bother!
        self.setting_changed("session_name", name)
        self.mdns_update()
        return f"session name set to {name}"

    def _ws_from_args(self, *args):
        # converts the args to valid window ids,
        # then returns all the window sources for those wids
        if len(args) == 0 or len(args) == 1 and args[0] == "*":
            # default to all if unspecified:
            wids = tuple(self._id_to_window.keys())
        else:
            wids = []
            for x in args:
                try:
                    wid = int(x)
                except ValueError:
                    raise ControlError(f"invalid window id: {x!r}") from None
                if wid in self._id_to_window:
                    wids.append(wid)
                else:
                    log(f"window id {wid} does not exist")
        wss = []
        for csource in tuple(self._server_sources.values()):
            for wid in wids:
                ws = csource.window_sources.get(wid)
                window = self._id_to_window.get(wid)
                if window and ws:
                    wss.append(ws)
        return wss

    def _set_encoding_property(self, name: str, value, *wids) -> str:
        for ws in self._ws_from_args(*wids):
            fn = getattr(ws, "set_" + name.replace("-", "_"))  # ie: "set_quality"
            fn(value)
        # now also update the defaults:
        for csource in tuple(self._server_sources.values()):
            csource.default_encoding_options[name] = value
        return f"{name} set to {value}"

    def control_command_quality(self, quality: int, *wids) -> str:
        return self._set_encoding_property("quality", quality, *wids)

    def control_command_min_quality(self, min_quality: int, *wids) -> str:
        return self._set_encoding_property("min-quality", min_quality, *wids)

    def control_command_max_quality(self, max_quality: int, *wids) -> str:
        return self._set_encoding_property("max-quality", max_quality, *wids)

    def control_command_speed(self, speed: int, *wids) -> str:
        return self._set_encoding_property("speed", speed, *wids)

    def control_command_min_speed(self, min_speed: int, *wids) -> str:
        return self._set_encoding_property("min-speed", min_speed, *wids)

    def control_command_max_speed(self, max_speed: int, *wids) -> str:
        return self._set_encoding_property("max-speed", max_speed, *wids)

    def control_command_auto_refresh(self, auto_refresh, *wids) -> str:
        delay = int(float(auto_refresh) * 1000.0)  # ie: 0.5 -> 500 (milliseconds)
        for ws in self._ws_from_args(*wids):
            ws.set_auto_refresh_delay(auto_refresh)
        return f"auto-refresh delay set to {delay}ms for windows {wids}"

    def control_command_refresh(self, *wids) -> str:
        for ws in self._ws_from_args(*wids):
            ws.full_quality_refresh({})
        return f"refreshed windows {wids}"

    def control_command_scaling_control(self, scaling_control, *wids) -> str:
        for ws in tuple(self._ws_from_args(*wids)):
            ws.set_scaling_control(scaling_control)
            ws.refresh()
        return f"scaling-control set to {scaling_control} on windows {wids}"

    def control_command_scaling(self, scaling, *wids) -> str:
        for ws in tuple(self._ws_from_args(*wids)):
            ws.set_scaling(scaling)
            ws.refresh()
        return f"scaling set to {scaling} on windows {wids}"

    def control_command_encoding(self, encoding: str, *args) -> str:
        if encoding in ("add", "remove"):
            cmd = encoding
            assert len(args) > 0
            encoding = args[0]
            wids = args[1:]
            for ws in tuple(self._ws_from_args(*wids)):
                encodings = list(ws.encodings)
                core_encodings = list(ws.core_encodings)
                for enc_list in (encodings, core_encodings):
                    if cmd == "add" and encoding not in enc_list:
                        log(f"adding {encoding} to {enc_list} for {ws}")
                        enc_list.append(encoding)
                    elif cmd == "remove" and encoding in enc_list:
                        log(f"removing {encoding} to {enc_list} for {ws}")
                        enc_list.remove(encoding)
                    else:
                        continue
                ws.encodings = tuple(encodings)
                ws.core_encodings = tuple(core_encodings)
                ws.do_set_client_properties(typedict())
                ws.refresh()
            return ["removed", "added"][cmd == "add"] + " " + encoding

        strict = None  # means no change
        if encoding in ("strict", "nostrict"):
            strict = encoding == "strict"
            encoding = args[0]
            wids = args[1:]
        elif len(args) > 0 and args[0] in ("strict", "nostrict"):
            # remove "strict" marker
            strict = args[0] == "strict"
            wids = args[1:]
        else:
            wids = args
        for ws in tuple(self._ws_from_args(*wids)):
            ws.set_new_encoding(encoding, strict)
            ws.refresh()
        return f"set encoding to {encoding}%s for windows {wids}" % ["", " (strict)"][int(strict or 0)]

    def control_command_request_update(self, encoding: str, geom, *args) -> str:
        wids = args
        now = monotonic()
        options = {
            "auto_refresh": True,
            "av-delay": 0,
        }
        log("request-update using %r, geometry=%s, windows(%s)=%s", encoding, geom, wids, self._ws_from_args(*wids))
        for ws in tuple(self._ws_from_args(*wids)):
            if geom == "all":
                x = y = 0
                w, h = ws.window_dimensions
            else:
                x, y, w, h = (int(x) for x in geom.split(","))
            ws.process_damage_region(now, x, y, w, h, encoding, options)
        return "damage requested"

    def control_command_clipboard_direction(self, direction: str, *_args) -> str:
        ch = self._clipboard_helper
        assert self.clipboard and ch
        direction = direction.lower()
        DIRECTIONS = ("to-server", "to-client", "both", "disabled")
        if direction not in DIRECTIONS:
            raise ValueError(f"invalid direction {direction!r}, must be one of " + csv(DIRECTIONS))
        self.clipboard_direction = direction
        can_send = direction in ("to-server", "both")
        can_receive = direction in ("to-client", "both")
        ch.set_direction(can_send, can_receive)
        msg = f"clipboard direction set to {direction!r}"
        log(msg)
        self.setting_changed("clipboard-direction", direction)
        return msg

    def control_command_clipboard_limits(self, max_send: int, max_recv: int, *_args) -> str:
        ch = self._clipboard_helper
        assert self.clipboard and ch
        ch.set_limits(max_send, max_recv)
        msg = f"clipboard send limit set to {max_send}, recv limit set to {max_recv} (single copy/paste)"
        log(msg)
        self.setting_changed("clipboard-limits", {'send': max_send, 'recv': max_recv})
        return msg

    def _control_video_subregions_from_wid(self, wid: int) -> list:
        if wid not in self._id_to_window:
            raise ControlError(f"invalid window {wid}")
        video_subregions = []
        for ws in self._ws_from_args(wid):
            vs = getattr(ws, "video_subregion", None)
            if not vs:
                log.warn(f"Warning: cannot set video region enabled flag on window {wid}")
                log.warn(f" no video subregion attribute found in {type(ws)}")
                continue
            video_subregions.append(vs)
        # log("_control_video_subregions_from_wid(%s)=%s", wid, video_subregions)
        return video_subregions

    def control_command_video_region_enabled(self, wid: int, enabled: bool) -> str:
        for vs in self._control_video_subregions_from_wid(wid):
            vs.set_enabled(enabled)
        return "video region %s for window %i" % (["disabled", "enabled"][int(enabled)], wid)

    def control_command_video_region_detection(self, wid: int, detection) -> str:
        for vs in self._control_video_subregions_from_wid(wid):
            vs.set_detection(detection)
        return "video region detection %s for window %i" % (["disabled", "enabled"][int(detection)], wid)

    def control_command_video_region(self, wid, x: int, y: int, w: int, h: int) -> str:
        for vs in self._control_video_subregions_from_wid(wid):
            vs.set_region(x, y, w, h)
        return "video region set to %s for window %i" % ((x, y, w, h), wid)

    def control_command_video_region_exclusion_zones(self, wid: int, zones) -> str:
        for vs in self._control_video_subregions_from_wid(wid):
            vs.set_exclusion_zones(zones)
        return f"video exclusion zones set to {zones} for window {wid}"

    def control_command_reset_video_region(self, wid: int) -> str:
        for vs in self._control_video_subregions_from_wid(wid):
            vs.reset()
        return f"reset video region heuristics for window {wid}"

    def control_command_lock_batch_delay(self, wid: int, delay: int) -> None:
        for ws in self._ws_from_args(wid):
            ws.lock_batch_delay(delay)

    def control_command_unlock_batch_delay(self, wid: int) -> None:
        for ws in self._ws_from_args(wid):
            ws.unlock_batch_delay()

    def control_command_set_lock(self, lock) -> str:
        self.lock = str_to_bool(lock)
        self.setting_changed("lock", lock is not False)
        self.setting_changed("lock-toggle", lock is None)
        return f"lock set to {self.lock}"

    def control_command_set_sharing(self, sharing) -> str:
        sharing = str_to_bool(sharing)
        message = f"sharing set to {self.sharing}"
        if sharing == self.sharing:
            return message
        self.sharing = sharing
        self.setting_changed("sharing", sharing is not False)
        self.setting_changed("sharing-toggle", sharing is None)
        if not sharing:
            # there can only be one ui client now,
            # disconnect all but the first ui_client:
            # (using the 'counter' value to figure out who was first connected)
            ui_clients = {
                getattr(ss, "counter", 0): proto
                for proto, ss in tuple(self._server_sources.items()) if getattr(ss, "ui_client", False)
            }
            n = len(ui_clients)
            if n > 1:
                for c in sorted(ui_clients)[1:]:
                    proto = ui_clients[c]
                    self.disconnect_client(proto, ConnectionMessage.SESSION_BUSY, "this session is no longer shared")
                message += f", disconnected {n - 1} clients"
        return message

    def control_command_set_ui_driver(self, uuid) -> str:
        ss = [s for s in self._server_sources.values() if s.uuid == uuid]
        if not ss:
            return f"source not found for uuid {uuid!r}"
        if len(ss) > 1:
            return f"more than one source found for uuid {uuid!r}"
        self.set_ui_driver(ss)
        return f"ui-driver set to {ss}"

    def control_command_key(self, keycode_str: str, press) -> str:
        if self.readonly:
            return "command key denied by readonly mode"
        try:
            if keycode_str.startswith("0x"):
                keycode = int(keycode_str, 16)
            else:
                keycode = int(keycode_str)
        except ValueError:
            raise ControlError(f"invalid keycode specified: {keycode_str!r} (not a number)") from None
        if keycode <= 0 or keycode >= 255:
            raise ControlError(f"invalid keycode value: {keycode} (must be between 1 and 255)")
        if press is not True:
            if press in ("1", "press"):
                press = True
            elif press in ("0", "unpress"):
                press = False
            else:
                raise ControlError("if present, the press argument must be one of: " + csv(
                    ("1", "press", "0", "unpress")))
        self.fake_key(keycode, press)
        return ("pressed" if press else "unpressed") + f" {keycode}"

    def control_command_audio_output(self, *args) -> str:
        msg = []
        for csource in tuple(self._server_sources.values()):
            msg.append(f"{csource} : " + str(csource.audio_control(*args)))
        return csv(msg)

    def control_command_workspace(self, wid: int, workspace: int) -> str:
        window = self._id_to_window.get(wid)
        if not window:
            raise ControlError(f"window {wid} does not exist")
        if "workspace" not in window.get_property_names():
            raise ControlError(f"cannot set workspace on window {window}")
        if workspace < 0:
            raise ControlError(f"invalid workspace value: {workspace}")
        window.set_property("workspace", workspace)
        return f"window {wid} moved to workspace {workspace}"

    def control_command_close(self, wid: int) -> str:
        window = self._id_to_window.get(wid)
        if not window:
            raise ControlError(f"window {wid} does not exist")
        window.request_close()
        return f"requested window {window} closed"

    def control_command_delete(self, wid: int) -> str:
        window = self._id_to_window.get(wid)
        if not window:
            raise ControlError(f"window {wid} does not exist")
        window.send_delete()
        return f"requested window {window} deleted"

    def control_command_move(self, wid: int, x: int, y: int) -> str:
        window = self._id_to_window.get(wid)
        if not window:
            raise ControlError(f"window {wid} does not exist")
        ww, wh = window.get_dimensions()
        count = 0
        for source in tuple(self._server_sources.values()):
            move_resize_window = getattr(source, "move_resize_window", None)
            if move_resize_window:
                move_resize_window(wid, window, x, y, ww, wh)
                count += 1
        return f"window {wid} moved to {x},{y} for {count} clients"

    def control_command_resize(self, wid: int, w: int, h: int) -> str:
        window = self._id_to_window.get(wid)
        if not window:
            raise ControlError(f"window {wid} does not exist")
        count = 0
        for source in tuple(self._server_sources.values()):
            resize_window = getattr(source, "resize_window", None)
            if resize_window:
                resize_window(wid, window, w, h)
                count += 1
        return f"window {wid} resized to {w}x{h} for {count} clients"

    def control_command_moveresize(self, wid: int, x: int, y: int, w: int, h: int) -> str:
        window = self._id_to_window.get(wid)
        if not window:
            raise ControlError(f"window {wid} does not exist")
        count = 0
        for source in tuple(self._server_sources.values()):
            move_resize_window = getattr(source, "move_resize_window", None)
            if move_resize_window:
                move_resize_window(wid, window, x, y, w, h)
                count += 1
        return f"window {wid} moved to {x},{y} and resized to {w}x{h} for {count} clients"

    def _process_control_request(self, protocol, packet: PacketType) -> None:
        """ client sent a command request through its normal channel """
        assert len(packet) >= 2, "invalid command request packet (too small!)"
        # packet[0] = "control"
        # this may end up calling do_handle_command_request via the adapter
        code, msg = self.process_control_command(protocol, *packet[1:])
        log("command request returned: %s (%s)", code, msg)

    def init_packet_handlers(self) -> None:
        self.add_packets(f"{ServerBaseControlCommands.PREFIX}-request")
        self.add_legacy_alias("command_request", f"{ServerBaseControlCommands.PREFIX}-request")
